package net.starmiddleware.starws2.providers.ccm;

/**
 * <p>Title: CCMProvider2</p>
 * <p>Description: a bridge for SOAP client accessing CCM services.  </p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: StarMiddleware.net</p>
 * @author Wang Hongjun              flymanmail@163.com
 * @version 1.0
 */

import java.util.*;
import javax.xml.namespace.*;
import javax.xml.rpc.holders.IntHolder;

import org.apache.axis.*;
import org.apache.axis.components.logger.LogFactory;
import org.apache.axis.description.*;
import org.apache.axis.enum.*;
import org.apache.axis.handlers.soap.*;
import org.apache.axis.message.*;
import org.apache.axis.providers.*;
import org.apache.axis.session.*;
import org.apache.axis.soap.*;
import org.apache.axis.utils.*;
import org.apache.commons.logging.*;
import org.omg.CORBA.*;
import org.omg.CORBA.InterfaceDefPackage.*;
import STARCCM.client.Components.*;
import net.starmiddleware.starws2.description.*;
import net.starmiddleware.starws2.encoding.CORBA.*;
import net.starmiddleware.starws2.util.CORBA.*;
import java.util.ArrayList;

public class CCMProvider
    extends BasicProvider {

  private static Log log =
      LogFactory.getLog(CCMProvider.class.getName());

  private static Log entLog =
      LogFactory.getLog(Constants.ENTERPRISE_LOG_CATEGORY);


  protected static final String OPTION_REPOSITORYURL = "repositoryURL";
  protected static final String OPTION_HOMEFINDERURL = "homeFinderURL";
  protected static final String OPTION_COMPONENTNAME = "componentName";
  protected static final String OPTION_COMPONENTIFNAME = "interfaceName";
  protected static final String OPTION_HOMEIFNAME = "homeInterfaceName";
  protected static final String OPTION_FACETIFNAMES = "facetInterfaceNames";
  protected static final String OPTION_ORBCLASS = "ORBClass";
  protected static final String OPTION_ORBSINGLETONCLASS = "ORBSingletonClass";
  protected static final String OPTION_ALLOWEDMETHODS = "allowedMethods";
  protected static final String OPTION_SCOPE = "scope";


  protected org.omg.CORBA.ORB orb = null;
  protected Repository repository = null;
  protected HomeFinder homeFinder = null;

  protected Hashtable orbCache = new Hashtable();
  protected Hashtable repositoryCache = new Hashtable();
  protected Hashtable homeFinderCache = new Hashtable();

  protected boolean isTargetServiceComponent = true;
  protected String implInterfaceName = null;
  protected String methodName4FetchFacetObj = null;

  /**
   * Init Service Desc
   * @param service
   * @param msgContext
   * @throws org.apache.axis.AxisFault
   */
  public void initServiceDesc(SOAPService service, MessageContext msgContext) throws
      org.apache.axis.AxisFault {

    //
    //init ORB& Repository& HomeFinder Obj
    //
    orb = initORB(service);
    assertNotNull(orb, "noORB");
    repository = initRepository(service);
    assertNotNull(orb, "noInfRepository");
    homeFinder = initHomeFinder(service);
    assertNotNull(orb, "noHomeFinder");

    //
    //set ServiceDesc
    //
    ServiceDesc serviceDesc = service.getServiceDescription();
    if (serviceDesc.getAllowedMethods() == null && service != null) {
      String allowedMethods = getAllowedMethods(service);
      if (allowedMethods != null && !"*".equals(allowedMethods)) {
        ArrayList methodList = new ArrayList();
        StringTokenizer tokenizer = new StringTokenizer(allowedMethods, " ,");
        while (tokenizer.hasMoreTokens()) {
          methodList.add(tokenizer.nextToken());
        }
        serviceDesc.setAllowedMethods(methodList);
      }
    }
  }

  /**
   * Init orb
   * @param service
   * @return
   * @throws AxisFault
   */

  protected ORB initORB(SOAPService service) throws AxisFault {

    //
    //retrieve ORB classes and set ORB's properties
    //
    java.util.Properties props = System.getProperties();
    String orbcls = (String) service.getOption(OPTION_ORBCLASS);
    String orbsingletcls = (String) service.getOption(OPTION_ORBSINGLETONCLASS);
    props.put("org.omg.CORBA.ORBClass", orbcls);
    props.put("org.omg.CORBA.ORBSingletonClass", orbsingletcls);

    //
    //get ORB Obj
    //
    if (orbCache.containsKey(props)) {
      return (org.omg.CORBA.ORB) orbCache.get(props);
    }

    ORB orb = null;
    try {
      String[] args = new String[1];
      args[0] = "";
      orb = (org.omg.CORBA.ORB) ORB.init(args, props);

    }
    catch (Exception e) {
      throw new AxisFault(Messages.getMessage("initORBError"));
    }

    orbCache.put(props, orb);
    return orb;
  }

  /**
   * Init repository
   * @param service
   * @return
   * @throws AxisFault
   */
  protected Repository initRepository(SOAPService service) throws AxisFault {

    //
    //retrieve Rpository url
    //
    String repositoryURL = (String) service.getOption(OPTION_REPOSITORYURL);
    assertNotNull(orb, "noORB");
    if (  (repositoryURL == null) || repositoryURL.length() < 1) {
      throw new AxisFault(Messages.getMessage(
          "initRepositoryError"));
    }

    //
    //get Rpository Obj
    //
    if (repositoryCache.containsKey(repositoryURL)) {
      return (org.omg.CORBA.Repository) repositoryCache.get(repositoryURL);
    }

    org.omg.CORBA.Repository repository = null;
    try {
      org.omg.CORBA.Object obj = orb.string_to_object(repositoryURL);
      repository = RepositoryHelper.narrow(obj);
    }
    catch (Exception e) {
      throw new AxisFault(Messages.getMessage(
          "initRepositoryError"));
    }

    repositoryCache.put(repositoryURL, repository);
    return repository;
  }

  /**
   * Init homeFinder
   * @param service
   * @return
   * @throws AxisFault
   */
  protected HomeFinder initHomeFinder(SOAPService service) throws AxisFault {

    //
    //retrieve HomeFinder url
    //
    assertNotNull(orb, "noORB");
    String homeFinderURL = (String) service.getOption(OPTION_HOMEFINDERURL);
    if ( (homeFinderURL == null) ||homeFinderURL.length() < 1) {
      throw new AxisFault(Messages.getMessage(
          "initHomeFinderError"));
    }

    //
    //get HomeFinder Obj
    //
    if (homeFinderCache.containsKey(homeFinderURL)) {
      return (HomeFinder) homeFinderCache.get(homeFinderURL);
    }

    HomeFinder homeFinder = null;
    try {
      org.omg.CORBA.Object obj = orb.string_to_object(homeFinderURL);
      homeFinder = HomeFinderHelper.narrow(obj);
    }
    catch (Exception e) {
      throw new AxisFault(Messages.getMessage(
          "initHomeFinderError"));
    }

    homeFinderCache.put(homeFinderURL, homeFinder);
    return homeFinder;
  }

  protected RPCElement getSoapRPCBody(MessageContext msgContext) throws
      AxisFault {

    RPCElement body = null;

    Message reqMsg = msgContext.getRequestMessage();
    SOAPEnvelope reqEnv = (SOAPEnvelope) reqMsg.getSOAPEnvelope();

    Vector bodies = reqEnv.getBodyElements();

    // Find the first "root" body element, which is the RPC call.
    for (int bNum = 0; body == null && bNum < bodies.size(); bNum++) {
      // If this is a regular old SOAPBodyElement, and it's a root,
      // we're probably a non-wrapped doc/lit service.  In this case,
      // we deserialize the element, and create an RPCElement "wrapper"
      // around it which points to the correct method.
      // FIXME : There should be a cleaner way to do this...
      if (! (bodies.get(bNum)instanceof RPCElement)) {
        //////////////////to do///////////////
      }
      else {
        body = (RPCElement) bodies.get(bNum);
      }
    }

    // If we still didn't find anything, report no body error.
    if (body == null) {
      throw new AxisFault(Messages.getMessage("noBody00"));
    }

    return body;

  }

  /**
   * get method names to be published
   * @param service
   * @return
   */
  protected String getAllowedMethods(Handler service) {
    String val = (String) service.getOption(OPTION_ALLOWEDMETHODS);
    return val;
  }

  /**
   * @param msgContext
   * @throws org.apache.axis.AxisFault
   */
  public void invoke(MessageContext msgContext) throws org.apache.axis.
      AxisFault {

    //
    //new a CORBAServiceDesc Obj
    //
    SOAPService service = msgContext.getService();
    ServiceDesc serviceDesc = service.getServiceDescription();
    CORBAServiceDesc corbaServiceDesc = new CORBAServiceDesc(serviceDesc);
    service.setServiceDescription(corbaServiceDesc);
    // set repository property
    corbaServiceDesc.setProperty(CORBAServiceDesc.PROP_REPOSITORY, repository);

    //
    //get invoked method name from the soap body
    //
    RPCElement body = getSoapRPCBody(msgContext);
    String methodName = body.getMethodName();

    //retieve component interface name
    String interfaceName = (String) service.getOption(OPTION_COMPONENTIFNAME);
    Contained ifContained = repository.lookup(interfaceName);
    assertNotNull(ifContained,"noInterfaceDefined");
    InterfaceDef ifdef = InterfaceDefHelper.narrow(ifContained);

    //
    //look up for the interface name that contains the invoked method
    //
    //check the component interface for the invoked method
    Contained methodContained = ifdef.lookup(methodName);
    if (methodContained != null) {
      isTargetServiceComponent = true;
      implInterfaceName = interfaceName;
    }
    else {

      //get out the facet interface which provides the invoked method
      String facetInterfaceNames = (String) service.getOption(
          OPTION_FACETIFNAMES);
      String facetInterfaceName = null;
      TypeCode facetTC=null;

      StringTokenizer st = new StringTokenizer(facetInterfaceNames, " ");
      while (st.hasMoreTokens()) {
        facetInterfaceName = st.nextToken();
        Contained facetContained = repository.lookup(facetInterfaceName);
        assertNotNull(facetContained,"noInterfaceDefined");
        InterfaceDef facetdef = InterfaceDefHelper.narrow(facetContained);
        methodContained = facetdef.lookup(methodName);
        if (methodContained != null) {
          isTargetServiceComponent = false;
          implInterfaceName = facetInterfaceName;
          facetTC = facetdef.type();
          break;
        }
      }

      if(implInterfaceName == null||implInterfaceName.length() <1) {
          throw new AxisFault(Messages.getMessage("noInterfaceDefined"));
      }

      //
      //check the component interface for method name which return the facet interface type
      //
      FullInterfaceDescription ifFullDesc = ifdef.describe_interface();
      OperationDescription[] opDescs = ifFullDesc.operations;
      for (int i = 0; i < opDescs.length; i++) {
        TypeCode tc = opDescs[i].result;
        if (tc.equal(facetTC)) {
          methodName4FetchFacetObj = opDescs[i].name;
          break;
        }
      }

      if (methodName4FetchFacetObj == null||methodName4FetchFacetObj.length()<1) {
        return;
      }
    }

    //
    //load CorbaServiceDesc with the implInterfaceName
    //
    corbaServiceDesc.loadServiceDescByIR(implInterfaceName);

    Message reqMsg = msgContext.getRequestMessage();
    SOAPEnvelope reqEnv = (SOAPEnvelope) reqMsg.getSOAPEnvelope();

    Message resMsg = msgContext.getResponseMessage();
    SOAPEnvelope resEnv = null;
    if (resMsg == null) {
      resEnv = new SOAPEnvelope(msgContext.getSOAPConstants());
      resMsg = new Message(resEnv);
      msgContext.setResponseMessage(resMsg);
    }
    else {
      resEnv = (SOAPEnvelope) resMsg.getSOAPEnvelope();
    }

    IntHolder scope = new IntHolder();
    org.omg.CORBA.Object serviceObj = getServiceObject(msgContext, service,
        scope);
    assertNotNull(serviceObj, "noServiceObject");

    try {
      processMessage(msgContext, reqEnv, resEnv, serviceObj);
    }
    catch (Exception exp) {
      entLog.debug(Messages.getMessage("toAxisFault00"), exp);
      throw AxisFault.makeFault(exp);
    }
    finally {
      if (log.isInfoEnabled()) {
        log.info("Exit: CorbaDynamic::invoke (" + this +")");

      }
    }

  }

  /**
   * generate WSDL for the service
   * @param msgContext
   * @throws AxisFault
   */
  public void generateWSDL(MessageContext msgContext) throws AxisFault {
    //todo.
  }

  protected org.omg.CORBA.Object getServiceObject(MessageContext msgContext,
                                                  Handler service,
                                                  IntHolder scopeHolder) throws
      org.apache.axis.AxisFault {

    org.omg.CORBA.Object serviceObj = null;
    org.omg.CORBA.Object componentObj = getComponentObject(msgContext,
        service, scopeHolder);
    serviceObj = componentObj;

    if (!isTargetServiceComponent) {
      assertNotNull(componentObj,"noServiceObject");
      org.omg.CORBA.Object facetObj = getFacetObject(msgContext, service,
          componentObj, scopeHolder);
      serviceObj = facetObj;
    }

    return serviceObj;

  }

  protected org.omg.CORBA.Object getComponentObject(MessageContext msgContext,
      Handler service, IntHolder scopeHolder) throws AxisFault {

    // scope can be "Request", "Session", "Application"
    // (as with Apache SOAP)
    Scope scope = Scope.getScope( (String) service.getOption(OPTION_SCOPE),
                                 Scope.DEFAULT);
    scopeHolder.value = scope.getValue();

    String serviceName = msgContext.getService().getName();
    if (serviceName == null) {
      serviceName = msgContext.getService().toString();
    }

    String interfaceName = (String) service.getOption(OPTION_COMPONENTIFNAME);
    String sessionName = serviceName + interfaceName;

    org.omg.CORBA.Object componentObj = null;
    if (scope == Scope.REQUEST) {
      componentObj = getNewComponentObject(msgContext, service);
      return componentObj;

    }
    else if (scope == Scope.SESSION) {

      // look in incoming session
      Session session = msgContext.getSession();
      if (session != null) {
        // This part isn't thread safe...
        synchronized (session) {
          // store service objects in session, indexed by class name
          componentObj = (org.omg.CORBA.Object) session.get(sessionName);
          if (componentObj == null) {
            componentObj = getNewComponentObject(msgContext, service);
            session.set(sessionName, componentObj);
          }
          return componentObj;
        }
      }
      else {
        // was no incoming session, sigh, treat as request scope
        scopeHolder.value = Scope.DEFAULT.getValue();
        componentObj = getNewComponentObject(msgContext, service);
        return componentObj;
      }
    }
    else if (scope == Scope.APPLICATION) {
      // MUST be AxisEngine here!
      AxisEngine engine = msgContext.getAxisEngine();
      Session appSession = engine.getApplicationSession();
      if (appSession != null) {
        // This part isn't thread safe
        synchronized (appSession) {
          // store service objects in session, indexed by class name
          componentObj = (org.omg.CORBA.Object) appSession.get(sessionName);
          if (componentObj == null) {
            componentObj = getNewComponentObject(msgContext, service);
            appSession.set(sessionName, componentObj);

          }
          return componentObj;
        }
      }
      else {
        // was no application session, sigh, treat as request scope
        // FIXME : Should we bomb in this case?
        scopeHolder.value = Scope.DEFAULT.getValue();
        componentObj = getNewComponentObject(msgContext, service);
        return componentObj;
      }
    }
    else {
      // NOTREACHED
      return null;
    }
  }

  protected org.omg.CORBA.Object getNewComponentObject(MessageContext
      msgContext,
      Handler service) throws AxisFault {

    org.omg.CORBA.Object homeObj = getHomeObject(msgContext, service);

    org.omg.CORBA.Object componentObj = null;
    try {

      //request context
      Context ctx = orb.get_default_context();

      //operationName
      String opName = "create";

      //arguments
      String homeIFName = (String) service.getOption(OPTION_HOMEIFNAME);

      Contained homeIFContained = repository.lookup(homeIFName);
      InterfaceDef homeIFDef = InterfaceDefHelper.narrow(homeIFContained);

      Contained opContained = homeIFDef.lookup(opName);
      OperationDef opDef = OperationDefHelper.narrow(opContained);
      NVList argList = orb.create_operation_list( (org.omg.CORBA.Object) opDef);
      //NVList argList =orb.create_list(0);
      //short shvar=12;
      //Any inany = argList.item(0).value();
      //inany.insert_short(shvar);

      //return type
      Any retany = orb.create_any();
      NamedValue result = orb.create_named_value("result", retany, 0);

      //request
      Request request = homeObj._create_request(ctx, opName, argList, result);
      request.set_return_type(org.omg.CORBA.ObjectHelper.type());
      request.invoke();

      //get invoke result
      //retany=request.return_value();
      retany = result.value();
      componentObj = org.omg.CORBA.ObjectHelper.extract(retany);

    }
    catch (Exception ex) {
      //System.err.println("Failed to create Component Object");
      throw new AxisFault(Messages.getMessage(
          "Failed to create Component Object"));

    }

    return componentObj;

  }

  protected org.omg.CORBA.Object getHomeObject(MessageContext msgContext,
                                               Handler service) throws
      AxisFault {

    org.omg.CORBA.Object homeObj = null;

    String componentName = (String) service.getOption(OPTION_COMPONENTNAME);
    if (componentName != null && !componentName.equals("")) {
      try {
        homeObj = homeFinder.find_home_by_name(componentName);

      }
      catch (Exception ex) {
        //System.err.println(" Failed to get Component Home Object");
        throw new AxisFault(Messages.getMessage(
            " Failed to get Component Home Object"));
      }
    }
    else {

    }

    return homeObj;
  }


  protected org.omg.CORBA.Object getFacetObject(MessageContext msgContext,
                                                Handler service,
                                                org.omg.CORBA.Object
                                                componentObj,
                                                IntHolder scopeHolder) throws
      AxisFault {

    // scope can be "Request", "Session", "Application"
    // (as with Apache SOAP)
    Scope scope = Scope.getScope( (String) service.getOption(OPTION_SCOPE),
                                 Scope.DEFAULT);
    scopeHolder.value = scope.getValue();

    String serviceName = msgContext.getService().getName();
    if (serviceName == null) {
      serviceName = msgContext.getService().toString();
    }

    String interfaceName = implInterfaceName;
    String sessionName = serviceName + implInterfaceName;

    org.omg.CORBA.Object facetObj = null;
    if (scope == Scope.REQUEST) {
      facetObj = getNewFacetObject(msgContext, service, componentObj);
      return facetObj;

    }
    else if (scope == Scope.SESSION) {

      // look in incoming session
      Session session = msgContext.getSession();
      if (session != null) {
        // This part isn't thread safe...
        synchronized (session) {
          // store service objects in session, indexed by class name
          facetObj = (org.omg.CORBA.Object) session.get(serviceName);
          if (facetObj == null) {
            facetObj = getNewFacetObject(msgContext, service, componentObj);
            session.set(serviceName, facetObj);
          }
          return facetObj;
        }
      }
      else {
        // was no incoming session, sigh, treat as request scope
        scopeHolder.value = Scope.DEFAULT.getValue();
        facetObj = getNewFacetObject(msgContext, service, componentObj);
        return facetObj;
      }
    }
    else if (scope == Scope.APPLICATION) {
      // MUST be AxisEngine here!
      AxisEngine engine = msgContext.getAxisEngine();
      Session appSession = engine.getApplicationSession();
      if (appSession != null) {
        // This part isn't thread safe
        synchronized (appSession) {
          // store service objects in session, indexed by class name
          facetObj = (org.omg.CORBA.Object) appSession.get(serviceName);
          if (facetObj == null) {
            facetObj = getNewFacetObject(msgContext, service, componentObj);
            appSession.set(serviceName, facetObj);
          }
          return facetObj;
        }
      }
      else {
        // was no application session, sigh, treat as request scope
        // FIXME : Should we bomb in this case?
        scopeHolder.value = Scope.DEFAULT.getValue();
        facetObj = getNewFacetObject(msgContext, service, componentObj);
        return facetObj;
      }
    }
    else {
      // NOTREACHED
      return null;
    }
  }

  protected org.omg.CORBA.Object getNewFacetObject(MessageContext msgContext,
      Handler service, org.omg.CORBA.Object componentObj) throws AxisFault {

    org.omg.CORBA.Object facetObj = null;
    try {
      //request context
      Context ctx = orb.get_default_context();

      //operationName
      String opName = methodName4FetchFacetObj;

      //arguments
      String ifName = (String) service.getOption(
          OPTION_COMPONENTIFNAME);

      Contained ifContained = repository.lookup(ifName);
      InterfaceDef ifDef = InterfaceDefHelper.narrow(ifContained);

      Contained opContained = ifDef.lookup(opName);
      OperationDef opDef = OperationDefHelper.narrow(opContained);
      NVList argList = orb.create_operation_list( (org.omg.CORBA.Object) opDef);
      //NVList argList =orb.create_list(0);
      //short shvar=12;
      //Any inany = argList.item(0).value();
      //inany.insert_short(shvar);

      //return type
      Any retany = orb.create_any();
      NamedValue result = orb.create_named_value("result", retany, 0);

      //request
      Request request = componentObj._create_request(ctx, opName, argList,
          result);
      request.set_return_type(org.omg.CORBA.ObjectHelper.type());
      request.invoke();

      //get invoke result
      //retany=request.return_value();
      retany = result.value();
      facetObj = org.omg.CORBA.ObjectHelper.extract(retany);

    }
    catch (Exception ex) {
      //System.err.println("Failed to make New Facet Object");
      throw new AxisFault(Messages.getMessage("Failed to make New Facet Object"));

    }

    return facetObj;
  }

  /*
    <?xml version='1.0' encoding='UTF-8'?>
    <SOAP-ENV:Envelope xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/" xmlns:xsi="http://www.w3.org/1999/XMLSchema-instance" xmlns:xsd="http://www.w3.org/1999/XMLSchema">
    <SOAP-ENV:Body>
    <ns1:echouShort xmlns:ns1="urn:Test:TestApp/Primtive" SOAP-ENV:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
    <namesh xsi:type="xsd:short">12</namesh>
    </ns1:echouShort>
    </SOAP-ENV:Body>
    </SOAP-ENV:Envelope>
   */

  protected void processMessage(MessageContext msgContext,
                                SOAPEnvelope reqEnv,
                                SOAPEnvelope resEnv,
                                org.omg.CORBA.Object serviceRef) throws
      Exception {
    Convert baseConvert = new Convert();
    if (log.isInfoEnabled()) {
      log.info("Enter: CorbaDynamic.processMessage()");
    }

    // get SOAPService, ServiceDes and OperationDesc
    SOAPService service = msgContext.getService();
    CORBAServiceDesc serviceDesc = (CORBAServiceDesc) service.
        getServiceDescription();
    CORBAOperationDesc operation = (CORBAOperationDesc) msgContext.getOperation();

    // parse Bodies
    Vector bodies = reqEnv.getBodyElements();
    if (log.isInfoEnabled()) {
      log.info(Messages.getMessage("bodyElems00", "" + bodies.size()));
      log.info(Messages.getMessage("bodyIs00", "" + bodies.get(0)));
    }
    RPCElement body = null;
    // Find the first "root" body element, which is the RPC call.
    for (int bNum = 0; body == null && bNum < bodies.size(); bNum++) {
      // If this is a regular old SOAPBodyElement, and it's a root,
      // we're probably a non-wrapped doc/lit service.  In this case,
      // we deserialize the element, and create an RPCElement "wrapper"
      // around it which points to the correct method.
      // FIXME : There should be a cleaner way to do this...
      if (! (bodies.get(bNum)instanceof RPCElement)) {
        SOAPBodyElement bodyEl = (SOAPBodyElement) bodies.get(bNum);
        // igors: better check if bodyEl.getID() != null
        // to make sure this loop does not step on SOAP-ENC objects
        // that follow the parameters! FIXME?
        if (bodyEl.isRoot() && operation != null && bodyEl.getID() == null) {
          ParameterDesc param = operation.getParameter(bNum);
          // at least do not step on non-existent parameters!
          if (param != null) {
            //     Object val = bodyEl.getValueAsType(param.getTypeQName());
            //    body = new RPCElement("",
            //               operation.getName(),
            //                     new Object [] { val });
          }
        }
      }
      else {
        body = (RPCElement) bodies.get(bNum);
      }
    }

    // special case code for a document style operation with no
    // arguments (which is a strange thing to have, but whatever)
    if (body == null) {
      // throw an error if this isn't a document style service
      if (!serviceDesc.getStyle().equals(Style.DOCUMENT)) {
        throw new Exception(Messages.getMessage("noBody00"));
      }

      // look for a method in the service that has no arguments,
      // use the first one we find.
      ArrayList ops = serviceDesc.getOperations();
      for (Iterator iterator = ops.iterator(); iterator.hasNext(); ) {
        OperationDesc desc = (OperationDesc) iterator.next();
        if (desc.getNumInParams() == 0) {
          // found one with no parameters, use it
          msgContext.setOperation(desc);
          // create an empty element
          body = new RPCElement(desc.getName());
          // stop looking
          break;
        }
      }

      // If we still didn't find anything, report no body error.
      if (body == null) {
        throw new Exception(Messages.getMessage("noBody00"));
      }
    }

    // get method names
    String methodName = body.getMethodName();
    org.omg.CORBA.Request request = serviceRef._request(methodName);
    //shuichao--start
    //QName name=body.getType();
    //QName CorbaQName=new QName(name.getNamespaceURI(),"corba");
    //body.setType(CorbaQName);
    //shuichao--end

    Vector args = body.getParams();
    int numArgs = args.size();
    if (operation == null) {
      QName qname = new QName(body.getNamespaceURI(), body.getName());
      operation = (CORBAOperationDesc) serviceDesc.getOperationByElementQName(
          qname);
    }

    if (operation == null) {
      throw new AxisFault(Messages.getMessage("noSuchOperation",
                                              methodName));
    }

    // A place to keep track of the out params (INOUTs and OUTs)
    ArrayList outs = new ArrayList();
    CORBAParameterDesc paramDesc = null;
    TypeCode tc = null;
    ArrayList parameters = operation.getParameters();


    int j = 0;
    for (int i = 0; i < operation.getNumParams(); i++) {
      paramDesc = (CORBAParameterDesc) operation.getParameter(i);
      tc = paramDesc.getTypeCode();
      int ParameterMode = paramDesc.getMode();

      switch (paramDesc.getMode()) {
        case ParameterDesc.IN:
          RPCParam rpcParam = (RPCParam) args.get(j);
          j = j + 1;
          java.lang.Object value = rpcParam.getValue();
          if (value == null) {
            System.out.println("parma is null");
            Any any1 = request.add_in_arg();
            any1 = null;
          }
          else {
            Any t_any = null;
            try {
              if (baseConvert.isSimpleType(tc.kind().value())) {
                t_any = baseConvert.simple2Any(orb, tc.kind().value(), value);
              }
              else {
                t_any = baseConvert.complex2Any(orb, tc, value);
              }
            }
            catch (net.starmiddleware.starws2.util.CORBA.
                   IllegalArgumentException e) {
              throw e;
            }
            try {
              org.omg.CORBA.portable.InputStream in = ( (Any) t_any).
                  create_input_stream();
              TypeCode tcd = ( (Any) t_any).type();
              request.add_in_arg().read_value(in, tcd);
            }
            catch (Exception e) {
              throw new Exception(Messages.getMessage(
                  "Error was found at create request Any "));
            }
          }

          break;

        case ParameterDesc.INOUT:
          rpcParam = (RPCParam) args.get(j);
          value = rpcParam.getValue();
          j = j + 1;
          if (value == null) {
            System.out.println("parma is null");
            Any any1 = request.add_inout_arg();
            any1 = null;
          }
          else {
            Any t_any = null;
            try {
              if (baseConvert.isSimpleType(tc.kind().value())) {
                t_any = baseConvert.simple2Any(orb, tc.kind().value(), value);
              }
              else {
                t_any = baseConvert.complex2Any(orb, tc, value);
              }
            }
            catch (net.starmiddleware.starws2.util.CORBA.
                   IllegalArgumentException e) {
              throw e;
            }
            try {
              org.omg.CORBA.portable.InputStream in = ( (Any) t_any).
                  create_input_stream();
              TypeCode tcd = ( (Any) t_any).type();
              request.add_inout_arg().read_value(in, tcd);
            }
            catch (Exception e) {
              throw new Exception(Messages.getMessage(
                  "Error was found at create request Any "));
            }
          }
          break;
        case ParameterDesc.OUT:

          Any t = request.add_out_arg();
          t.type(tc);
          break;

      }

    }

    // See if any subclasses want a crack at faulting on a bad operation
    // FIXME : Does this make sense here???


    String allowedMethods = (String) service.getOption("allowedMethods");
    checkMethodName(msgContext, allowedMethods, operation.getName());

    //shucihao-start
    tc = operation.getReturnTypeCode();
    /*if (tc.kind().value()==TCKind._tk_alias){
        try {
            request.set_return_type(tc.content_type());
            tc=tc.content_type();
        }catch(Exception e) {}
              }
              else { request.set_return_type(tc);}
     */
    request.set_return_type(tc);
    //shuichao--edn
    System.out.println("CORBA DII invocation starting....");
    request.invoke();


    if (request.env().exception() != null) {
      throw (SystemException) request.env().exception();
    }
    //String CORBAInterfaceName = (String) service.getOption(OPTION_INFNAME);
    String CORBAInterfaceName = implInterfaceName;
    java.util.StringTokenizer buffer = new StringTokenizer(CORBAInterfaceName,
        "::");
    String modulName = buffer.nextToken();
    String interfaceName = buffer.nextToken();
    String className = modulName + "." + interfaceName;
    QName namespace = new QName("http://www.omg.org/IDL-Mapped/", className);
    baseConvert.setDefaultType(namespace);
    //shuichao-end
    NVList arguments = request.arguments();
    java.lang.Object obj = new java.lang.Object();
    for (int i = 0; i < arguments.count(); i++) {
      int flag = arguments.item(i).flags();
      Any any_out = arguments.item(i).value();
      if (flag != 1) {
        //    Any any_out = arguments.item(i).value();
        TypeCode tc_out = ( (CORBAParameterDesc) parameters.get(i)).getTypeCode();
        if (baseConvert.isSimpleType(tc_out.kind().value())) {
          obj = baseConvert.any2Simple(any_out, tc_out);
        }
        else {
          obj = baseConvert.any2Complex(orb, any_out, tc_out);
        }
        if (obj instanceof ComplexType) {
          ( (ComplexType) obj).setType(namespace);
        }
        RPCParam p = new RPCParam( ( (ParameterDesc) parameters.get(i)).
                                  getQName(), obj);

        p.setParamDesc( (ParameterDesc) parameters.get(i));
        outs.add(p);
      }
    }

    java.lang.Object objRes = null;
    Any res = request.return_value();
    if (baseConvert.isSimpleType(tc.kind().value())) {
      objRes = baseConvert.any2Simple(res, tc);
    }
    else {
      objRes = baseConvert.any2Complex(orb, res, tc);
    }
    if (objRes instanceof ComplexType) {
      ( (ComplexType) objRes).setType(namespace);
    }

    RPCElement resBody = new RPCElement(methodName + "Response");

    resBody.setPrefix(body.getPrefix());
    resBody.setNamespaceURI(body.getNamespaceURI());
    resBody.setEncodingStyle(msgContext.getEncodingStyle());

    QName returnQName = operation.getReturnQName();
    if (returnQName == null) {
      returnQName = new QName("", methodName + "Return");
    }

    // For SOAP 1.2, add a result
    if (msgContext.getSOAPConstants() == SOAPConstants.SOAP12_CONSTANTS) {
      RPCParam result = new RPCParam
          (Constants.QNAME_RPC_RESULT, returnQName.getLocalPart());
      resBody.addParam(result);
    }

    RPCParam param = new RPCParam(returnQName, objRes);
    param.setParamDesc(operation.getReturnParamDesc());
    resBody.addParam(param);

    if (!outs.isEmpty()) {
      for (Iterator i = outs.iterator(); i.hasNext(); ) {
        // We know this has a holder, so just unwrap the value
        param = (RPCParam) i.next();
        resBody.addParam(param);

      }
    }

    resEnv.addBodyElement(resBody);
  }



  protected void checkMethodName(MessageContext msgContext,
                                 String allowedMethods,
                                 String methodName) throws Exception {
    // Our version doesn't need to do anything, though inherited
    // ones might.
  }

  protected void assertNotNull(java.lang.Object obj, String str) throws
      AxisFault {
    if (obj == null) {
      throw new AxisFault(Messages.getMessage(str));
    }
  }

}
