/***************************************************************************
 * Title:        CorbaDynamic
 * Description:  The version is developed  at Oct,2003
 * Copyright:    Copyright (c) 2003
 * Company:      starwebservice
 * @author       CORBA WebService Group
 * @version      1.0
******************************************************************************/

package net.starmiddleware.starws2.providers.corba;

import java.util.*;
import javax.xml.namespace.*;
import javax.xml.rpc.holders.IntHolder;

import org.apache.axis.*;
import org.apache.axis.components.logger.LogFactory;
import org.apache.axis.description.*;
import org.apache.axis.encoding.*;
import org.apache.axis.enum.*;
import org.apache.axis.handlers.soap.*;
import org.apache.axis.message.*;
import org.apache.axis.providers.*;
import org.apache.axis.soap.*;
import org.apache.axis.utils.*;
import org.apache.axis.wsdl.fromJava.*;
import org.apache.commons.logging.*;
import org.omg.CORBA.*;
import org.w3c.dom.*;
import net.starmiddleware.starws2.description.*;
import net.starmiddleware.starws2.encoding.CORBA.*;
import net.starmiddleware.starws2.util.CORBA.*;


public class DynProvider extends BasicProvider {



  protected static Log log =
        LogFactory.getLog(DynProvider.class.getName());

     protected static Log entLog =
        LogFactory.getLog(Constants.ENTERPRISE_LOG_CATEGORY);

    public static final String OPTION_ORBCLASS="ORBClass";
    public static final String OPTION_ORBSingletonClass="ORBSingletonClass";
    public static final String OPTION_INFNAME = "interfaceName";
    public static final String OPTION_CORBANAME="corbaName";
    public static final String OPTION_CORBAURL="corbaUrl";
    public static final String OPTION_CORBAREF="corbaRef";
    public static final String OPTION_LOCATORURL="locatorUrl";
    public static final String OPTION_IRURL = "repositoryUrl";

    public static final String OPTION_ALLOWEDMETHODS = "allowedMethods";
   // public static final String OPTION_CLASSNAME = "className";
   // public static final String OPTION_IS_STATIC = "isStatic";
    public static final String OPTION_CLASSPATH = "classPath";
    public static final String OPTION_WSDL_PORTTYPE="wsdlPortType";
    public static final String OPTION_WSDL_SERVICEELEMENT="wsdlServiceElement";
    public static final String OPTION_WSDL_SERVICEPORT="wsdlServicePort";
    public static final String OPTION_WSDL_TARGETNAMESPACE="wsdlTargetNamespace";

    public static final String OPTION_SCOPE = "scope";


    private org.omg.CORBA.ORB  orb = null;
    private org.omg.CORBA.Repository  repository = null;
    private Hashtable repCache = new Hashtable();
    private Hashtable  orbCache = new Hashtable();




  public void initServiceDesc(SOAPService service, MessageContext msgContext) throws org.apache.axis.AxisFault {

        String infName = (String)service.getOption(OPTION_INFNAME) ;
        if (infName == null) {
            throw new AxisFault(Messages.getMessage("noInterfaceName"));
        }

        ServiceDesc serviceDescription = service.getServiceDescription();
        serviceDescription.setProperty(CORBAServiceDesc.PROP_INFNAME ,infName) ;
        CORBAServiceDesc CORBAServiceDescription=new CORBAServiceDesc(serviceDescription);
        CORBAServiceDescription.setProperty(CORBAServiceDesc.PROP_INFNAME ,infName) ;


        // And the allowed methods, if necessary
        if (serviceDescription.getAllowedMethods() == null && service != null) {
            String allowedMethods = getAllowedMethods(service);
            if (allowedMethods != null && !"*".equals(allowedMethods)) {
                ArrayList methodList = new ArrayList();
                StringTokenizer tokenizer = new StringTokenizer(allowedMethods, " ,");
                while (tokenizer.hasMoreTokens()) {
                    methodList.add(tokenizer.nextToken());
                }
                serviceDescription.setAllowedMethods(methodList);
                CORBAServiceDescription.setAllowedMethods(methodList);
            }
        }

        orb = initORB(service);
        repository = initRepository(service);
        //serviceDescription.setProperty(ServiceDesc.PROP_REPOSITORY ,repository) ;
        //serviceDescription.loadServiceDescByIR(infName);
        CORBAServiceDescription.setProperty(CORBAServiceDesc.PROP_REPOSITORY ,repository) ;
        CORBAServiceDescription.loadServiceDescByIR(infName);
        service.setServiceDescription(CORBAServiceDescription);

    }


    private org.omg.CORBA.Repository  initRepository(SOAPService service)throws AxisFault {

         org.omg.CORBA.Object obj = null;
         org.omg.CORBA.Repository rep = null;

        String ir=(String)service.getOption(OPTION_IRURL);

        if ((orb == null) || (ir == null))
              return null;
        if (repCache.containsKey(ir) )
         return (org.omg.CORBA.Repository )repCache.get(ir) ;

         try
         {
          obj = orb.string_to_object(ir);
          rep = org.omg.CORBA.RepositoryHelper.narrow(obj);
          }catch(Exception e ){
           // System.err.println("error occur in resolve Repository IOR");
            throw new AxisFault( Messages.getMessage("no Inf Repository ") );
          }

          repCache.put(ir,rep);
        return rep;
    }

    private org.omg.CORBA.ORB  initORB(SOAPService service) throws AxisFault {
     org.omg.CORBA.ORB  orb = null;
     java.util.Properties props=null;
     try{
        props = System.getProperties();
        String orbcls = (String)service.getOption(OPTION_ORBCLASS) ;
        String orbsingletcls=(String)service.getOption(OPTION_ORBSingletonClass);

        props.put("org.omg.CORBA.ORBClass", orbcls);
        props.put("org.omg.CORBA.ORBSingletonClass",orbsingletcls);

        if (orbCache.containsKey(orbsingletcls) )
       return (org.omg.CORBA.ORB )orbCache.get(props) ;

        String[] args = new String[1];
        args[0] = "";
        orb=(org.omg.CORBA.ORB)ORB.init(args,props);

         } catch(Exception e ) {
            throw new AxisFault(Messages.getMessage("init Orb Fail") );
        }

      orbCache.put(props,orb);
      return (org.omg.CORBA.ORB )orb;
    }


  private String getAllowedMethods(Handler service)
 {
    String val = (String)service.getOption(OPTION_ALLOWEDMETHODS);
    return val;
 }


  protected void checkMethodName(MessageContext msgContext,
                                  String allowedMethods,
                                  String methodName)
       throws Exception
   {
       // Our version doesn't need to do anything, though inherited
       // ones might.
   }


  public void invoke(MessageContext msgContext) throws org.apache.axis.AxisFault
  {

    if (log.isInfoEnabled())
      log.info("Enter: CorbaDynamic::invoke (" + this +")");

    String serviceName = msgContext.getTargetService();
    Handler service = msgContext.getService();

    IntHolder scope = new IntHolder();
    org.omg.CORBA.Object serviceRef = null;

    try {

      serviceRef = getServiceReference(msgContext, service, scope);

      Message resMsg = msgContext.getResponseMessage();
      SOAPEnvelope resEnv;

      if (resMsg == null) {
        resEnv = new SOAPEnvelope(msgContext.getSOAPConstants());
        resMsg = new Message(resEnv);
        msgContext.setResponseMessage(resMsg);
      }
      else {
        resEnv = (SOAPEnvelope) resMsg.getSOAPEnvelope();
      }
      Message reqMsg = msgContext.getRequestMessage();
      SOAPEnvelope reqEnv = (SOAPEnvelope) reqMsg.getSOAPEnvelope();
      processMessage(msgContext, reqEnv, resEnv, serviceRef);
    }
    catch (Exception exp) {
      entLog.debug(Messages.getMessage("toAxisFault00"), exp);
      throw AxisFault.makeFault(exp);
    }
    finally {
      // If this is a request scoped service object which implements
      // ServiceLifecycle, let it know that it's being destroyed now.
      /*      if (serviceRef != null &&
          scope.value == Scope.REQUEST.getValue() &&
          serviceRef instanceof ServiceLifecycle) {
        ( (ServiceLifecycle) serviceRef).destroy();
        }
             }
       */
      if (log.isInfoEnabled())
        log.info("Exit: CorbaDynamic::invoke (" + this +")");

      }
    }


    public void generateWSDL(MessageContext msgContext) throws AxisFault {
        if (log.isInfoEnabled())
            log.info("Enter: DynProvider::generateWSDL (" + this + ")");

      /* Find the service we're invoking so we can grab it's options */
        /***************************************************************/
        SOAPService service = msgContext.getService();
        CORBAServiceDesc serviceDesc = (CORBAServiceDesc)service.getInitializedServiceDesc(msgContext);

        // Calculate the appropriate namespaces for the WSDL we're going
        // to put out.
        //
        // If we've been explicitly told which namespaces to use, respect
        // that.  If not:
        //
        // The "interface namespace" should be either:
        // 1) The namespace of the ServiceDesc
        // 2) The transport URL (if there's no ServiceDesc ns)

        try {
            // Location URL is whatever is explicitly set in the MC
            String locationUrl =
                    msgContext.getStrProp(MessageContext.WSDLGEN_SERV_LOC_URL);

            if (locationUrl == null) {
                // If nothing, try what's explicitly set in the ServiceDesc
                locationUrl = serviceDesc.getEndpointURL();
            }

            if (locationUrl == null) {
                // If nothing, use the actual transport URL
                locationUrl = msgContext.getStrProp(MessageContext.TRANS_URL);
            }

            // Interface namespace is whatever is explicitly set
            String interfaceNamespace =
                    msgContext.getStrProp(MessageContext.WSDLGEN_INTFNAMESPACE);

            if (interfaceNamespace == null) {
                // If nothing, use the default namespace of the ServiceDesc
                interfaceNamespace = serviceDesc.getDefaultNamespace();
            }

            if (interfaceNamespace == null) {
                // If nothing still, use the location URL determined above
                interfaceNamespace = locationUrl;
            }

            net.starmiddleware.starws2.wsdl.fromIDL.Emitter emitter = new net.starmiddleware.starws2.wsdl.fromIDL.Emitter();

            // service alias may be provided if exact naming is required,
            // otherwise Axis will name it according to the implementing class name
            //String alias = (String)service.getOption("alias");
            //if(alias != null) emitter.setServiceElementName(alias);

            Style style = serviceDesc.getStyle();
            if (style == Style.RPC) {
                emitter.setMode(Emitter.MODE_RPC);
            } else if (style == Style.DOCUMENT) {
                emitter.setMode(Emitter.MODE_DOCUMENT);
            } else if (style == Style.WRAPPED) {
                emitter.setMode(Emitter.MODE_DOC_WRAPPED);
            }

            // If a wsdl target namespace was provided, use the targetNamespace.
            // Otherwise use the interfaceNamespace constructed above.
            String targetNamespace = (String) service.getOption(OPTION_WSDL_TARGETNAMESPACE);
            if (targetNamespace == null ||
                targetNamespace.length() == 0) {
                targetNamespace = interfaceNamespace;
            }
            emitter.setIntfNamespace(targetNamespace);

            emitter.setLocationUrl(locationUrl);
            emitter.setServiceDesc(serviceDesc);
            emitter.setTypeMapping((TypeMapping)msgContext.getTypeMappingRegistry().
                                   getTypeMapping(Constants.URI_DEFAULT_SOAP_ENC));
            emitter.setDefaultTypeMapping((TypeMapping)msgContext.getTypeMappingRegistry().
                                          getDefaultTypeMapping());

            String wsdlPortType = (String) service.getOption(OPTION_WSDL_PORTTYPE);
            String wsdlServiceElement = (String) service.getOption(OPTION_WSDL_SERVICEELEMENT);
            String wsdlServicePort = (String) service.getOption(OPTION_WSDL_SERVICEPORT);

            if (wsdlPortType != null && wsdlPortType.length() > 0) {
                emitter.setPortTypeName(wsdlPortType);
            }
            if (wsdlServiceElement != null && wsdlServiceElement.length() > 0) {
                emitter.setServiceElementName(wsdlServiceElement);
            }
            if (wsdlServicePort != null && wsdlServicePort.length() > 0) {
                emitter.setServicePortName(wsdlServicePort);
            }

            Document  doc = emitter.emit(Emitter.MODE_ALL);

            msgContext.setProperty("WSDL", doc);
        } catch (NoClassDefFoundError e) {
            entLog.info( Messages.getMessage("toAxisFault00"), e );
            throw new AxisFault(e.toString(), e);
        } catch (Exception e) {
            entLog.info( Messages.getMessage("toAxisFault00"), e );
            throw AxisFault.makeFault(e);
        }

        if (log.isInfoEnabled())
            log.info("Exit: DynProvider::generateWSDL (" + this + ")");

    }


  public org.omg.CORBA.Object getServiceReference (MessageContext msgContext,
                                  Handler service,
                                  IntHolder scopeHolder)throws org.apache.axis.AxisFault
                              {
      org.omg.CORBA.Object  obj=null;
     String  corbaUrl=(String)service.getOption(OPTION_CORBAURL);
     return  orb.string_to_object(corbaUrl);

  }


  protected void processMessage (MessageContext msgContext,
                               SOAPEnvelope reqEnv,
                               SOAPEnvelope resEnv,
                               org.omg.CORBA.Object serviceRef)
       throws Exception  {
    Convert baseConvert=new Convert();
     if (log.isInfoEnabled()) {
          log.info("Enter: CorbaDynamic.processMessage()");
      }

           SOAPService service = msgContext.getService();
           CORBAServiceDesc serviceDesc =(CORBAServiceDesc) service.getServiceDescription();
           CORBAOperationDesc operation =(CORBAOperationDesc) msgContext.getOperation();

              Vector    bodies = reqEnv.getBodyElements();
                  if (log.isInfoEnabled()) {
                   log.info(Messages.getMessage("bodyElems00", "" + bodies.size()));
                   log.info(Messages.getMessage("bodyIs00", "" + bodies.get(0)));
                 }
             RPCElement   body = null;
    // Find the first "root" body element, which is the RPC call.
       for ( int bNum = 0 ; body == null && bNum < bodies.size() ; bNum++ ) {
           // If this is a regular old SOAPBodyElement, and it's a root,
           // we're probably a non-wrapped doc/lit service.  In this case,
           // we deserialize the element, and create an RPCElement "wrapper"
           // around it which points to the correct method.
           // FIXME : There should be a cleaner way to do this...
           if (!(bodies.get(bNum) instanceof RPCElement)) {
                SOAPBodyElement bodyEl = (SOAPBodyElement)bodies.get(bNum);
              // igors: better check if bodyEl.getID() != null
              // to make sure this loop does not step on SOAP-ENC objects
              // that follow the parameters! FIXME?
           if (bodyEl.isRoot() && operation != null && bodyEl.getID() == null) {
                 ParameterDesc param = operation.getParameter(bNum);
                // at least do not step on non-existent parameters!
                 if(param != null) {
              //     Object val = bodyEl.getValueAsType(param.getTypeQName());
               //    body = new RPCElement("",
                 //               operation.getName(),
                   //                     new Object [] { val });
                     }
                 }
             } else {
                 body = (RPCElement) bodies.get( bNum );
             }
         }

         // special case code for a document style operation with no
         // arguments (which is a strange thing to have, but whatever)
         if (body == null) {
             // throw an error if this isn't a document style service
             if (!serviceDesc.getStyle().equals(Style.DOCUMENT)) {
                 throw new Exception(Messages.getMessage("noBody00"));
             }

             // look for a method in the service that has no arguments,
             // use the first one we find.
             ArrayList ops = serviceDesc.getOperations();
             for (Iterator iterator = ops.iterator(); iterator.hasNext();) {
                 OperationDesc desc = (OperationDesc) iterator.next();
                 if (desc.getNumInParams() == 0) {
                     // found one with no parameters, use it
                     msgContext.setOperation(desc);
                     // create an empty element
                     body = new RPCElement(desc.getName());
                     // stop looking
                     break;
                 }
             }

             // If we still didn't find anything, report no body error.
             if (body == null) {
                 throw new Exception(Messages.getMessage("noBody00"));
             }
         }

         String methodName = body.getMethodName();
         org.omg.CORBA.Request request=serviceRef._request(methodName);
         //shuichao--start
         //QName name=body.getType();
         //QName CorbaQName=new QName(name.getNamespaceURI(),"corba");
         //body.setType(CorbaQName);
         //shuichao--end

         Vector args = body.getParams();
        int numArgs = args.size();
         if (operation == null) {
             QName qname = new QName(body.getNamespaceURI(), body.getName());
             operation =(CORBAOperationDesc) serviceDesc.getOperationByElementQName(qname);
         }

        if (operation == null) {
                             throw new AxisFault(Messages.getMessage("noSuchOperation",
                                                                      methodName));
                         }

         // A place to keep track of the out params (INOUTs and OUTs)
         ArrayList outs = new ArrayList();
         CORBAParameterDesc paramDesc=null;
         TypeCode tc=null;
         ArrayList parameters = operation.getParameters();

        int j=0;
        for (int i=0;i<operation.getNumParams();i++) {
           paramDesc =(CORBAParameterDesc) operation.getParameter(i);
           tc = paramDesc.getTypeCode();
           int ParameterMode = paramDesc.getMode();

           switch (paramDesc.getMode())
           {
             case ParameterDesc.IN:
                RPCParam rpcParam = (RPCParam) args.get(j);
                j=j+1;
                java.lang.Object value = rpcParam.getValue();
                if (value == null) {
                      System.out.println("parma is null");
                      Any any1 = request.add_in_arg();
                     any1 = null;
                   }
                  else
                     {  Any t_any=null;
                        try {
                          if ( baseConvert.isSimpleType(tc.kind().value()))
                            { t_any = baseConvert.simple2Any(orb, tc.kind().value(), value); }
                           else {  t_any = baseConvert.complex2Any(orb, tc, value); }
                               } catch (net.starmiddleware.starws2.util.CORBA.IllegalArgumentException e) {
                                  throw e; }
                           try { org.omg.CORBA.portable.InputStream in=((Any)t_any).create_input_stream();//Any value
                               TypeCode tcd = ( (Any) t_any).type();
                               request.add_in_arg().read_value(in, tcd);
                             }
                             catch (Exception e) {
                               throw new Exception(Messages.getMessage("Error was found at create request Any "));
                              }
                       }

               break;

            case ParameterDesc.INOUT:
                 rpcParam = (RPCParam) args.get(j);
                 value = rpcParam.getValue();
                 j=j+1;
                if (value == null) {
                     System.out.println("parma is null");
                     Any any1 = request.add_inout_arg();
                    any1 = null;
                   }
                 else
                    {  Any t_any=null;
                       try {
                         if ( baseConvert.isSimpleType(tc.kind().value()))
                           { t_any = baseConvert.simple2Any(orb, tc.kind().value(), value); }
                          else {  t_any = baseConvert.complex2Any(orb, tc, value); }
                              } catch (net.starmiddleware.starws2.util.CORBA.IllegalArgumentException e) {
                                 throw e; }
                          try { org.omg.CORBA.portable.InputStream in=((Any)t_any).create_input_stream();
                              TypeCode tcd = ( (Any) t_any).type();
                              request.add_inout_arg().read_value(in, tcd);
                            }
                            catch (Exception e) {
                              throw new Exception(Messages.getMessage("Error was found at create request Any "));
                             }
                      }
                     break;
              case ParameterDesc.OUT:

                Any t=request.add_out_arg();
                t.type(tc);
                break;

 }


           }



         // See if any subclasses want a crack at faulting on a bad operation
         // FIXME : Does this make sense here???


         String allowedMethods = (String)service.getOption("allowedMethods");
         checkMethodName(msgContext, allowedMethods, operation.getName());

         tc=operation.getReturnTypeCode();
         /*if (tc.kind().value()==TCKind._tk_alias){
             try {
                 request.set_return_type(tc.content_type());
                 tc=tc.content_type();
             }catch(Exception e) {}
         }
         else { request.set_return_type(tc);}
        */
         request.set_return_type(tc);
         //shuichao--edn
         log.debug("CORBA DII invocation starting....");

         request.invoke();


         if(request.env().exception() != null)
             throw  (SystemException)request.env().exception();

         String CORBAInterfaceName=(String) service.getOption(OPTION_INFNAME);
         java.util.StringTokenizer buffer=new StringTokenizer(CORBAInterfaceName,"::");
         String modulName="";
         if(buffer.hasMoreTokens()) modulName=buffer.nextToken();
         String interfaceName=buffer.nextToken();
         String className=modulName+"."+interfaceName;
         QName  namespace=new QName("http://www.omg.org/IDL-Mapped/",className);
         baseConvert.setDefaultType(namespace);

         NVList arguments = request.arguments();
        java.lang.Object obj = new java.lang.Object();
         for (int i = 0; i< arguments.count(); i++){
             int  flag = arguments.item(i).flags();
              Any any_out = arguments.item(i).value();
             if ( flag !=1){
             //    Any any_out = arguments.item(i).value();
                 TypeCode tc_out = ((CORBAParameterDesc)parameters.get(i)).getTypeCode();
                 if(baseConvert.isSimpleType(tc_out.kind().value())){
                   obj=baseConvert.any2Simple(any_out,tc_out);
                 }else{
                   obj=baseConvert.any2Complex(orb,any_out,tc_out);
                 }
                 if(obj instanceof ComplexType){
                    ((ComplexType)obj).setType(namespace);
                 }
                 RPCParam p = new RPCParam(((ParameterDesc)parameters.get(i)).getQName(), obj);

                 p.setParamDesc((ParameterDesc)parameters.get(i));
                 outs.add(p);
             }
         }

         java.lang.Object objRes=null;
         Any res=request.return_value();
         if(baseConvert.isSimpleType(tc.kind().value())){
           objRes=baseConvert.any2Simple(res,tc);
         }
         else{
           objRes=baseConvert.any2Complex(orb,res,tc);
         }
         if(objRes instanceof ComplexType){
            ((ComplexType)objRes).setType(namespace);
           }

         RPCElement resBody = new RPCElement(methodName + "Response");

         resBody.setPrefix( body.getPrefix() );
         resBody.setNamespaceURI( body.getNamespaceURI() );
         resBody.setEncodingStyle(msgContext.getEncodingStyle());

                   QName returnQName = operation.getReturnQName();
                   if (returnQName == null) {
                       returnQName = new QName("", methodName + "Return");
                   }

                   // For SOAP 1.2, add a result
                   if (msgContext.getSOAPConstants() == SOAPConstants.SOAP12_CONSTANTS)
                   {
                       RPCParam result = new RPCParam
                          (Constants.QNAME_RPC_RESULT, returnQName.getLocalPart());
                       resBody.addParam(result);
                   }

                   RPCParam param = new RPCParam(returnQName, objRes);
                   param.setParamDesc(operation.getReturnParamDesc());
                   resBody.addParam(param);


              if (!outs.isEmpty()) {
             for (Iterator i = outs.iterator(); i.hasNext();) {
                 // We know this has a holder, so just unwrap the value
                 param = (RPCParam) i.next();
                 resBody.addParam(param);

             }
         }

         resEnv.addBodyElement(resBody);
     }
 }
