package net.starmiddleware.starws2.providers.jms;

import org.apache.axis.description.*;
import org.apache.axis.providers.java.MsgProvider;
import org.apache.axis.providers.BasicProvider;
import org.apache.axis.MessageContext;
import org.apache.axis.AxisFault;
import org.apache.axis.i18n.Messages;
import org.apache.axis.description.OperationDesc;
import org.apache.axis.message.SOAPBodyElement;
import org.apache.axis.message.SOAPEnvelope;
import org.w3c.dom.Element;
import org.w3c.dom.Document;
import org.apache.axis.handlers.soap.SOAPService;
import org.apache.axis.providers.java.JavaProvider;
import org.apache.axis.AxisEngine;
import org.apache.axis.utils.cache.ClassCache;
import org.apache.axis.utils.cache.JavaClass;
import org.apache.axis.utils.ClassUtils;
import java.lang.reflect.Method;
import java.util.Vector;
import org.apache.axis.Handler;
import org.apache.axis.Message;
import javax.xml.rpc.holders.IntHolder;
import org.apache.axis.enum.Scope;
import net.starmiddleware.starws2.providers.jms.JMSConstants;

import java.util.ArrayList;
import java.util.StringTokenizer;
/**
 * <p>Title: </p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: </p>
 * @author not attributable
 * @version 1.0
 */

public class JMSProvider extends BasicProvider {

  public static final String OPTION_JNDI_PROPERTIES = "jndiproperties";
  public static final String OPTION_JMS_DESITINATIONNAME = "JMSDestinationName";
  public static final String OPTION_JMS_MESSAGETPYE = "JMSMessageType";
  public static final String OPTION_ALLOWEDMETHODS = "allowedMethods";
  public static final String OPTION_CLASSNAME = "className";
  public static final String MSG_METHOD_SOAPENVELOPE = "processSoapEnvelopeMessageService";
  public static final String MSG_METHOD_ELEMENTARRAY = "processElementsMessageService";
  public static final String MSG_METHOD_DOCUMENT = "processDocumentMessageSerivce";
  public static final String MSG_METHOD_BODYARRAY = "processSoapBodyElementMessageService";
  public String serviceName = null;
  public Object serviceObj = null;
  public int methodType ;


  public JMSProvider() {
  }

  public void initServiceDesc(SOAPService service, MessageContext msgContext)
      throws org.apache.axis.AxisFault {
    ServiceDesc serviceDescription = service.getServiceDescription();

    // And the allowed methods, if necessary
    if (serviceDescription.getAllowedMethods() == null && service != null) {
        String allowedMethods = getAllowedMethods(service);
        if (allowedMethods != null && !"*".equals(allowedMethods)) {
            ArrayList methodList = new ArrayList();
            StringTokenizer tokenizer = new StringTokenizer(allowedMethods, " ,");
            while (tokenizer.hasMoreTokens()) {
                methodList.add(tokenizer.nextToken());
            }
            serviceDescription.setAllowedMethods(methodList);
        }
      }
  }

  /**
   * Return the option in the configuration that contains the service class
   * name
   */
  protected String getServiceClassNameOptionName() {
    return OPTION_CLASSNAME;
  }

  protected String getJNDIPropertiesName() {
    return OPTION_JNDI_PROPERTIES;
  }

  protected String getJMSDestination() {
    return OPTION_JMS_DESITINATIONNAME;
  }

  protected String getJMSMessageType() {
    return OPTION_JMS_MESSAGETPYE;
  }


  public void invoke(MessageContext msgContext) throws org.apache.axis.AxisFault{

    try {

      serviceName = msgContext.getService().getName();
      Handler service = msgContext.getService();
      IntHolder scope = new IntHolder();
      ////
      //only four kinds of JMS Services in the SOAP2JMSBridge, so use hard coding here
      //create new service object according to the serviceName
      ////
      if(serviceName.equals(JMSProvider.MSG_METHOD_SOAPENVELOPE)){
        serviceObj = new processSoapEnvelopeMessageService();
        methodType = 1;
      }else if (serviceName.equals(JMSProvider.MSG_METHOD_BODYARRAY)){
        serviceObj = new processSoapBodyElementMessageService();
        methodType = 2;
      }else if (serviceName.equals(JMSProvider.MSG_METHOD_ELEMENTARRAY)){
        serviceObj = new processElementsMessageService();
        methodType = 3;
      }else if (serviceName.equals(JMSProvider.MSG_METHOD_DOCUMENT)){
        serviceObj = new processDocumentMessageSerivce();
        methodType = 4;
      }

      //췵resEnvMessage
      Message resMsg = msgContext.getResponseMessage();
      SOAPEnvelope resEnv;
      if (resMsg == null) {
        resEnv = new SOAPEnvelope(msgContext.getSOAPConstants());
        resMsg = new Message(resEnv);
        msgContext.setResponseMessage(resMsg);
      }
      else {
        resEnv = (SOAPEnvelope) resMsg.getSOAPEnvelope();
      }

      Message reqMsg = msgContext.getRequestMessage();
      SOAPEnvelope reqEnv = (SOAPEnvelope) reqMsg.getSOAPEnvelope();
      processMessage(msgContext, reqEnv, resEnv, serviceObj);

    }
    catch (Exception exp) {
      throw AxisFault.makeFault(exp);
    }
    finally {
      // If this is a request scoped service object which implements
      // ServiceLifecycle, let it know that it's being destroyed now.
      /*      if (serviceObj != null &&
          scope.value == Scope.REQUEST.getValue() &&
          serviceObj instanceof ServiceLifecycle) {
        ( (ServiceLifecycle) serviceObj).destroy();
        }
             }
       */
      if (log.isInfoEnabled()) {
        log.info("Exit: JMSProvider::invoke (" + this +")");

      }
    }

  }

  private String getAllowedMethods(Handler service)
 {
    String val = (String)service.getOption(OPTION_ALLOWEDMETHODS);
    return val;
 }


  private void processSOAPJMSInfo(MessageContext msgContext){

    //JNDIProperties file name
    String jndiFileName  = (String)msgContext.getService().getOption(OPTION_JNDI_PROPERTIES);
    String s = System.getProperty("starws.home");
    if (jndiFileName == null || jndiFileName.length() == 0) {
      // Try the old option for backwards-compatibility
      // added by zk be careful!
      jndiFileName = JMSConstants.DEFAULT_JNDI_PROPERTIES;
    }
    JMSConstants.setJNDIFileName(jndiFileName);

    //JMSdestination
    String JMSdestination  = (String)msgContext.getService().getOption(OPTION_JMS_DESITINATIONNAME);
    if (JMSdestination == null || JMSdestination.length() == 0) {
      // Try the old option for backwards-compatibility
      // added by zk be careful!
      JMSdestination = JMSConstants.DEFAULT_JMSDestination;
    }
    JMSConstants.setDestinationName(JMSdestination);

    //JMSDestinationName
    String JMSType  = (String)msgContext.getService().getOption(OPTION_JMS_MESSAGETPYE);
    if (JMSType == null || JMSType.length() == 0) {
      // Try the old option for backwards-compatibility
      // added by zk be careful!
      JMSType = JMSConstants.DEFAULT_JMSMessageType;
    }
    JMSConstants.setJMSConfigMessageType(JMSType);
  }


  public void processMessage(MessageContext msgContext,
                             SOAPEnvelope reqEnv,
                             SOAPEnvelope resEnv,
                             Object obj) throws Exception {
    //added by zk
    processSOAPJMSInfo(msgContext);

    if (serviceName != JMSProvider.MSG_METHOD_SOAPENVELOPE) {
      // dig out just the body, and pass it on
      Vector bodies = reqEnv.getBodyElements();
      Object argObjects[] = new Object[1];

      switch (methodType) {
        // SOAPBodyElement [] / SOAPBodyElement []
        case 2:
          SOAPBodyElement[] bodyElements =
              new SOAPBodyElement[bodies.size()];
          bodies.toArray(bodyElements);
          argObjects[0] = bodyElements;
          SOAPBodyElement[] bodyResult =
              (SOAPBodyElement[])((processSoapBodyElementMessageService)serviceObj).processSoapBodyElement(bodyElements);
          if (bodyResult != null) {
            for (int i = 0; i < bodyResult.length; i++) {
              SOAPBodyElement bodyElement = bodyResult[i];
              resEnv.addBodyElement(bodyElement);
            }
          }
          return;

          // Element [] / Element []
        case 3:
          Element[] elements = new Element[bodies.size()];
          for (int i = 0; i < elements.length; i++) {
            SOAPBodyElement body = (SOAPBodyElement) bodies.get(i);
            elements[i] = body.getAsDOM();
          }
          argObjects[0] = elements;
          Element[] elemResult =
              (Element[]) ((processElementsMessageService)serviceObj).processElements(elements);
          if (elemResult != null) {
            for (int i = 0; i < elemResult.length; i++) {
              if (elemResult[i] != null)
                resEnv.addBodyElement(
                    new SOAPBodyElement(elemResult[i]));
            }
          }
          return;

          // Element [] / Element []
        case 4:
          Document doc = ((SOAPBodyElement) bodies.get(0)).getAsDocument();
          argObjects[0] = doc;
          Document resultDoc =
              (Document) ((processDocumentMessageSerivce)serviceObj).processDocument(doc);
          if (resultDoc != null) {
            resEnv.addBodyElement(new SOAPBodyElement(
                resultDoc.getDocumentElement()));
          }
          return;
      }
    }
    else {
      SOAPEnvelope argObjects[] = new SOAPEnvelope[2];
      // SOAPEnvelope / SOAPEnvelope
      argObjects[0] = reqEnv;
      argObjects[1] = resEnv;
      ((processSoapEnvelopeMessageService)serviceObj).processsoapEnvelope(reqEnv,resEnv);
      return;
    }

    // SHOULD NEVER GET HERE...
    throw new AxisFault(Messages.getMessage("badMsgMethodStyle"));
  }


}