package net.starmiddleware.starws2.security.rbac;

import org.apache.axis.handlers.BasicHandler;
import org.apache.axis.*;
import org.apache.axis.handlers.BasicHandler;
import org.apache.axis.*;
import org.apache.axis.message.*;
import java.io.IOException;
import java.io.InputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.*;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.cert.CertificateException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.crypto.BadPaddingException;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.dom.*;
import javax.xml.transform.*;
import javax.xml.transform.stream.*;

import com.ibm.dom.util.DOMUtil;
import com.ibm.xml.dsig.IDResolver;
import com.ibm.xml.dsig.XSignatureException;
import com.ibm.xml.enc.AlgorithmFactoryExtn;
import com.ibm.xml.enc.EncryptedKeyRetriever;
import com.ibm.xml.enc.EncryptionContext;
import com.ibm.xml.enc.EncryptionEngine;
import com.ibm.xml.enc.DecryptionContext;
import com.ibm.xml.enc.KeyInfoResolver;
import com.ibm.xml.enc.KeyInfoResolvingException;
import com.ibm.xml.enc.StructureException;
import com.ibm.xml.enc.type.EncryptedData;
import com.ibm.xml.enc.util.AdHocIdResolver;
import com.ibm.xml.enc.util.KeyStoreKeyInfoResolver;
import com.ibm.xml.enc.util.SameDocumentEncryptedKeyRetriever;

import org.apache.xerces.parsers.DOMParser;
import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.XMLSerializer;
import org.apache.xpath.XPathAPI;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

//import com.ibm.crypto.provider.IBMJCE;

/**
 * <p>Title: </p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2002</p>
 * <p>Company: </p>
 * @author zm
 * @version 1.0
 */

public class EncryptionHandler
    extends BasicHandler {

    //necessary parameters in .wsdd file

    static String KEY_INFO_FILE = "keyinfoFile";

    static String SECURITY_PROVIDER = "securityProvider";

    static String ENCRYPTION_TEMPLATE = "templateFile";

    //unnecessary parameters in .wsdd file

    static String NAMESPACE_URI = "namespaceURI";

    static String LOCAL_NAME = "localName";

    //implement the BasicHandler's abstract method
    public void invoke(MessageContext msgContext) throws org.apache.axis.
        AxisFault {

        try {
            String relativeRootPath = RBACUtil.getSystemRootPath();
                //(String) getOption("relativeRootPath");


            //prepare the security provider
            AlgorithmFactoryExtn af = new AlgorithmFactoryExtn();
            if (!"*".equals( (String) getOption(SECURITY_PROVIDER))) {
                af.setProvider( (String) getOption(SECURITY_PROVIDER));
            }


            //prepare the keystore and keys's info
            Keyinfo ki = getKeyinfo( relativeRootPath + (String) getOption(KEY_INFO_FILE));

            //get the Document to be encrypted
            Document d = msgContext.getCurrentMessage().getSOAPEnvelope().
                getAsDocument();


            KeyInfoResolver kir = getKeyInfoResolver(ki, af, null, null);
            kir.setOperationMode(KeyInfoResolver.ENCRYPT_MODE);

            EncryptionContext ec = new EncryptionContext();
            ec.setAlgorithmFactory(af);
            ec.setKeyInfoResolver(kir);


            //get Body's sub-element
            MessageElement elem = (MessageElement) msgContext.getCurrentMessage()
                .getSOAPEnvelope().getBody().getChildElements().next();

            //get the specified "namespaceURI"
            String namespaceURI = (String) getOption(NAMESPACE_URI);
            if (namespaceURI == null) {
                namespaceURI = elem.getNamespaceURI();
            }

            //get the specified "localName"
            String localName = (String) getOption(LOCAL_NAME);
            if (localName == null) {
                localName = elem.getName();
            }

            //localize the element to be encrypted
            String s = "//*[namespace-uri()='" + namespaceURI +
                "' and local-name()='" + localName + "']";

            NodeList nl = XPathAPI.selectNodeList(d, s);

            //set the encryption framework template
            Element e = getDocument( relativeRootPath + (String) getOption(ENCRYPTION_TEMPLATE)).
                getDocumentElement();

            //encrypt
            encryptElements(nl, ec, e);

            printDocument(d);

            //substitute the old SOAP message
            Transformer t = TransformerFactory.newInstance().newTransformer();
            Writer w = new StringWriter();
            t.transform(new DOMSource(d), new StreamResult(w));
            SOAPEnvelope newSe = new SOAPEnvelope(
                new ByteArrayInputStream(w.toString().getBytes()));
            msgContext.setCurrentMessage(new Message(newSe));

        }
        catch (Exception e) {
            throw AxisFault.makeFault(e);
        }
    }

    private static void encryptElements(NodeList nodes,
                                        EncryptionContext context,
                                        Element template) {
        for (int i = 0, l = nodes.getLength(); i < l; i++) {
            Node n = nodes.item(i);
            if (n.getNodeType() == Node.ELEMENT_NODE) {
                try {
                    encryptElement( (Element) n, context, template);

                }
                catch (Exception e) {
                    printStackTrace(e);
                }
            }
        }
    }

    private static void encryptElement(Element elem, EncryptionContext context,
                                       Element template) throws
        BadPaddingException, IOException, IllegalBlockSizeException,
        InvalidAlgorithmParameterException, InvalidKeyException,
        KeyInfoResolvingException, NoSuchAlgorithmException,
        NoSuchPaddingException,
        NoSuchProviderException, StructureException {
        context.setData(elem);
        context.setEncryptedType( (Element) template.cloneNode(true), null, null, null);
        context.setKey(null);
        context.encrypt();
        context.replace();
    }

    static Keyinfo getKeyinfo(String name) throws IOException, SAXException {
        Element e = getDocument(name).getDocumentElement();

        return new Keyinfo(e);
    }

    static Document getDocument(String name) throws IOException, SAXException {
        DOMParser dp = new DOMParser();
        dp.setErrorHandler(new ErrorHandler() {
            //
            // Methods of interface: ErrorHandler
            //

            public void warning(SAXParseException exception) throws
                SAXException {
                printStackTrace(exception);
            }

            public void error(SAXParseException exception) throws SAXException {
                printStackTrace(exception);
            }

            public void fatalError(SAXParseException exception) throws
                SAXException {
                throw exception;
            }
        });
        dp.setFeature("http://apache.org/xml/features/dom/defer-node-expansion", false);
        dp.parse(name);

        return dp.getDocument();
    }

    static KeyInfoResolver getKeyInfoResolver(Keyinfo keyinfo,
                                              AlgorithmFactoryExtn factory,
                                              IDResolver idRes,
                                              EncryptedKeyRetriever retriever) throws
        CertificateException, IOException, KeyStoreException,
        NoSuchAlgorithmException,
        NoSuchProviderException {
        return getKeyInfoResolver(keyinfo, factory, idRes, retriever,
                                  getKeyStore(keyinfo));
    }

    static KeyInfoResolver getKeyInfoResolver(Keyinfo keyinfo,
                                              AlgorithmFactoryExtn factory,
                                              IDResolver idRes,
                                              EncryptedKeyRetriever retriever,
                                              KeyStore store) {
        KeyStoreKeyInfoResolver kiRes = new KeyStoreKeyInfoResolver(store);
        kiRes.setAlgorithmFactory(factory);
        kiRes.setIdResolver(idRes);
        kiRes.setEncryptedKeyRetriever(retriever);
        for (Iterator i = keyinfo.fAlias2Keypass.keySet().iterator(); i.hasNext(); ) {
            String s = (String) i.next();
            kiRes.putAliasAndPassword(s, (char[]) keyinfo.fAlias2Keypass.get(s));
        }

        return kiRes;
    }

    static KeyStore getKeyStore(Keyinfo keyinfo) throws CertificateException,
        IOException, KeyStoreException, NoSuchAlgorithmException,
        NoSuchProviderException {
        String s = keyinfo.fStoretype;
        if (s == null) {
            s = KeyStore.getDefaultType();
        }
        KeyStore store = null;
        if (keyinfo.fProvider != null) {
            store = KeyStore.getInstance(s, keyinfo.fProvider);

        }
        else {
            store = KeyStore.getInstance(s);
        }

        s = keyinfo.fStorename;
        if (s == null) {
            s = System.getProperty("user.home") + File.separator + ".keystore";
        }
        File f = new File(s);
        InputStream is = null;
        if (f.exists()) {
            is = new FileInputStream(f);
        }
        store.load(is, keyinfo.fStorepass);

        return store;
    }

    static void printStackTrace(Throwable thr) {
        if (thr instanceof SAXException) {
            Throwable t = ( (SAXException) thr).getException();
            if (t != null) {
                thr = t;
            }
        }
        else if (thr instanceof TransformerException) {
            Throwable t = ( (TransformerException) thr).getException();
            if (t != null) {
                thr = t;
            }
        }
        else if (thr instanceof XSignatureException) {
            Throwable t = ( (XSignatureException) thr).getException();
            if (t != null) {
                thr = t;
            }
        }
        thr.printStackTrace();
    }

    static void printDocument(Document doc) throws IOException {
        OutputFormat of = new OutputFormat(doc);
        of.setPreserveSpace(true);
        XMLSerializer xs = new XMLSerializer(System.out, of);
        xs.serialize(doc);
        System.out.println();
    }


    static class Keyinfo {
        //
// Data
        //

        String fStorename;
        String fStoretype;
        char[] fStorepass;
        String fProvider;
        Map fAlias2Keypass;

        private String alias;
        private char[] pass;

        //
        // Constructors
        //

        Keyinfo(Element keyinfo) {
            if (!isOfType(keyinfo)) {
                throw new IllegalArgumentException("Not keyinfo element");
            }

            fAlias2Keypass = new HashMap();
            init(keyinfo);
        }

        //
        // Methods
        //

        static boolean isOfType(Element elem) {
            return (elem != null && "keyinfo".equals(elem.getTagName()));
        }

        private void init(Node node) {
            for (Node n = node.getFirstChild(); n != null; n = n.getNextSibling()) {
                init(n);
            }

            if (node.getNodeType() == Node.ELEMENT_NODE) {
                String s = node.getNodeName();
                if (s.equals("alias")) {
                    alias = DOMUtil.getStringValue(node);

                }
                else if (s.equals("key")) {
                    fAlias2Keypass.put(alias, pass);
                    alias = null;
                    pass = null;

                }
                else if (s.equals("keystore")) {
                    fStorepass = pass;
                    pass = null;

                }
                else if (s.equals("name")) {
                    fStorename = RBACUtil.getSystemRootPath() + DOMUtil.getStringValue(node);

                }
                else if (s.equals("password")) {
                    pass = DOMUtil.getStringValue(node).toCharArray();

                }
                else if (s.equals("provider")) {
                    fProvider = DOMUtil.getStringValue(node);

                }
                else if (s.equals("type")) {
                    fStoretype = DOMUtil.getStringValue(node);
                }
            }
        }
    }

}
