package net.starmiddleware.starws2.security.rbac;

import java.net.URI;
import java.net.URISyntaxException;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import com.sun.xacml.*;
import com.sun.xacml.attr.*;
import com.sun.xacml.combine.*;
import com.sun.xacml.cond.*;
import com.sun.xacml.ctx.*;
import com.sun.xacml.finder.*;
import com.sun.xacml.finder.impl.*;

import org.w3c.dom.*;
import java.io.*;

import org.apache.axis.*;

/**
 * <p>Title: </p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: </p>
 * @author zm
 * @version 1.0
 */


public class PermissionManager {


  public static String POLICY_REPOSITORY = RBACUtil.getSystemRootPath() +
      //System.getProperty("user.dir") +
      "//conf/security/rbac/policy_repository/";


  private static String DEFAULT_POLICY_FILE = "default.xml";

  private static String log4Permission = RBACUtil.getSystemRootPath() +
      // System.getProperty("user.dir") +
      "//conf/security/rbac/logs/Permission.log";


  public static String assignPermission(String role, String service,
                                        String method) throws AxisFault {
    try {

      String policyFile = POLICY_REPOSITORY + role + ".xml";
      File file = new File(policyFile);
      String result = null;


      if (service.equals("") || method.equals("")) {
        result =
            "Service name or method name maybe empty, please assign them first.";
      }
      else {

        if (! (file.exists())) {
          PolicyProducer.makePolicyFile(role, service, method);
        }

        else {

          FileInputStream fis = new FileInputStream(file);
          Node node = parseInput(fis, "Policy");
          fis.close();

          Policy policy = Policy.getInstance(node);
          List rules = policy.getChildren();

          rules = addRule(rules, service, method);

          Policy newPolicy = new Policy(policy.getId(),
                                        (RuleCombiningAlgorithm) (
              policy.
              getCombiningAlg()),
                                        policy.getDescription(),
                                        policy.getTarget(), rules);

          newPolicy.encode(System.out);

          FileOutputStream fos = new FileOutputStream(file);

          newPolicy.encode(fos);
          fos.close();

        }

        result = "For Role: " + role +
            " You have succeeded in assigning permissions for Service: " +
            service + "'s method " + method;

        RBACUtil.log4String(log4Permission, result);
      }

      return result;
    }
    catch (Exception e) {
      throw AxisFault.makeFault(e);
      //e.printStackTrace();
    }

  }

  //
  public static String deassignPermission(String role, String service,
                                          String method) throws AxisFault {
    try {

      String policyFile = POLICY_REPOSITORY + role + ".xml";
      File file = new File(policyFile);
      String result = null;


      if (service.equals("") || method.equals("")) {
        result =
            "Service name or method name maybe empty, please assign them first.";
      }
      else {


        if (! (file.exists())) {
          throw new Exception("no policy file exists");
        }

        else {

          FileInputStream fis = new FileInputStream(file);
          Node node = parseInput(fis, "Policy");
          fis.close();

          Policy policy = Policy.getInstance(node);
          List rules = policy.getChildren();

          rules = deleteRule(rules, service, method);
          Policy newPolicy = new Policy(policy.getId(),
                                        (RuleCombiningAlgorithm) (
              policy.
              getCombiningAlg()),
                                        policy.getDescription(),
                                        policy.getTarget(), rules);

          newPolicy.encode(System.out);

          FileOutputStream fos = new FileOutputStream(file);

          newPolicy.encode(fos);
          fos.close();

        }

        result = "For Role: " + role +
            " You have succeeded in assigning permissions for Service: " +
            service + "'s method " + method;

        RBACUtil.log4String(log4Permission, result);
      }

      return result;

    }
    catch (Exception e) {
      throw AxisFault.makeFault(e);
      //e.printStackTrace();
    }
  }


  private static List deleteRule(List ruleList, String service, String method) throws
      Exception {

    Iterator ruleIterator = ruleList.iterator();
    boolean rule_exist = false;
    Rule existRule = null;

    List newRuleList = new ArrayList();

    while (ruleIterator.hasNext()) {
      Rule rule = (Rule) (ruleIterator.next());
      if (rule.getId().toString().equals(service)) {
        rule_exist = true;
        existRule = rule;
      }
      else {
        newRuleList.add(rule);
      }
    }

    if (rule_exist) {
      Target target = existRule.getTarget();
      List actions = target.getActions();
      List newActions = deleteAction(actions, method);
      if (!newActions.isEmpty()) {
        Target newTarget = new Target(target.getSubjects(),
                                      target.getResources(),
                                      newActions);
        Rule newRule = new Rule(existRule.getId(), existRule.getEffect(),
                                existRule.getDescription(),
                                newTarget, existRule.getCondition());
        newRuleList.add(newRule);
      }
      return newRuleList;
    }
    else {
      System.out.println(
          "**************************the service doesn't exist**********************");
      return newRuleList;
    }
  }


  private static List deleteAction(List actions, String method) throws
      Exception {


    List newActions = new ArrayList();
    Iterator actionIterator = actions.iterator();
    boolean action_exist = false;

    while (actionIterator.hasNext()) {
      List action = (List) (actionIterator.next());
      Iterator matchIterator = action.iterator();
      List tempAction = new ArrayList();

      while (matchIterator.hasNext()) {
        TargetMatch tm = (TargetMatch) (matchIterator.next());
        StringAttribute sa = (StringAttribute) (tm.getMatchValue());
        if (sa.getValue().equals(method)) {
          System.out.println(
              "**************************the method already exists**********************");
          action_exist = true;
        }
        else {
          tempAction.add(tm);
        }
      }

      if (! (tempAction.isEmpty())) {
        newActions.add(tempAction);
      }

    }

    if (!action_exist) {
      System.out.println(
          "**************************the method doesn't exist**********************");
    }

    return newActions;
  }


  private static Document getDocument(InputStream input) throws
      ParsingException {
    try {
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      factory.setIgnoringComments(true);

      DocumentBuilder builder = null;

      // we're not validating
      factory.setNamespaceAware(false);
      factory.setValidating(false);

      builder = factory.newDocumentBuilder();

      Document doc = builder.parse(input);
      return doc;
    }
    catch (Exception e) {
      throw new ParsingException("Error tring to parse " + e);
    }

  }

  private static Node parseInput(InputStream input, String rootTag) throws
      ParsingException {
    NodeList nodes = null;

    try {
      Document doc = getDocument(input);
      nodes = doc.getElementsByTagName(rootTag);
    }
    catch (Exception e) {
      throw new ParsingException("Error tring to parse " + rootTag +
                                 "Type", e);
    }

    if (nodes.getLength() != 1)
      throw new ParsingException("Only one " + rootTag + "Type allowed "
                                 + "at the root of a Context doc");

    return nodes.item(0);
  }


  private static List addAction(List actions, String method) throws Exception {

    List newActions = new ArrayList();

    Iterator actionIterator = actions.iterator();
    while (actionIterator.hasNext()) {
      List action = (List) (actionIterator.next());
      Iterator matchIterator = action.iterator();
      List tempAction = new ArrayList();

      while (matchIterator.hasNext()) {
        TargetMatch tm = (TargetMatch) (matchIterator.next());
        StringAttribute sa = (StringAttribute) (tm.getMatchValue());
        if (sa.getValue().equals(method)) {
          System.out.println(
              "**************************the method already exists**********************");
          return actions;
        }
        else {
          tempAction.add(tm);
        }
      }

      newActions.add(tempAction);
    }

    List newAction = PolicyProducer.createAction(method);
    newActions.add(newAction);
    return newActions;
  }

  private static TargetMatch createTargetMatchForAction(String method) {
    try {
      String actionMatchId = EqualFunction.NAME_STRING_EQUAL;

      URI actionDesignatorType = new URI(StringAttribute.identifier);

      URI actionDesignatorId = new URI("method");

      AttributeDesignator actionDesignator =
          new AttributeDesignator(AttributeDesignator.ACTION_TARGET,
                                  actionDesignatorType,
                                  actionDesignatorId, false);

      StringAttribute actionValue = new StringAttribute(method);

      // get the factory that handles Target functions and get an
      // instance of the right function
      FunctionFactory factory = FunctionFactory.getTargetInstance();
      Function function = factory.createFunction(actionMatchId);

      // create the TargetMatch
      return new TargetMatch(TargetMatch.ACTION, function,
                             actionDesignator, actionValue);
    }
    catch (Exception e) {
      // note that in this example, we should never hit this case, but
      // in the real world you need to worry about exceptions, especially
      // from the factory
      return null;
    }
  }


  private static List addRule(List ruleList, String service, String method) throws
      Exception {

    Iterator ruleIterator = ruleList.iterator();
    boolean rule_exist = false;
    List newRuleList = new ArrayList();
    Rule existRule = null;

    while (ruleIterator.hasNext()) {
      Rule rule = (Rule) (ruleIterator.next());
      if (rule.getId().toString().equals(service)) {
        rule_exist = true;
        existRule = rule;
      }
      else {
        newRuleList.add(rule);
      }
    }

    if (rule_exist) {
      Target target = existRule.getTarget();
      List actions = target.getActions();
      List newActions = addAction(actions, method);
      Target newTarget = new Target(target.getSubjects(),
                                    target.getResources(), newActions);
      Rule newRule = new Rule(existRule.getId(), existRule.getEffect(),
                              existRule.getDescription(),
                              newTarget, existRule.getCondition());
      newRuleList.add(newRule);
      return newRuleList;
    }

    else {
      Rule newRule = PolicyProducer.createRule(service, method);
      newRuleList.add(newRule);
      return newRuleList;
    }

  }


  public static Node listAssignedPermissions(String role) throws AxisFault {
    try {


      System.out.println(POLICY_REPOSITORY);

      String policyFile = POLICY_REPOSITORY + role + ".xml";
      File file = new File(policyFile);

      if (! (file.exists())) {
        String defaultPolicyFile = POLICY_REPOSITORY +
            DEFAULT_POLICY_FILE;
        File defaultFile = new File(defaultPolicyFile);
        FileInputStream fis = new FileInputStream(defaultFile);
        Document doc = getDocument(fis);
        fis.close();
        Element roleElement = doc.createElement("role");
        roleElement.setAttribute("name", role);
        return roleElement;
      }

      else {
        FileInputStream fis = new FileInputStream(file);
        Node node = parseInput(fis, "Policy");
        fis.close();
        fis = new FileInputStream(file);
        Document doc = getDocument(fis);
        fis.close();
        Element roleElement = doc.createElement("role");
        roleElement.setAttribute("name", role);

        Policy policy = Policy.getInstance(node);
        List rules = policy.getChildren();
        Iterator ruleIterator = rules.iterator();
        while (ruleIterator.hasNext()) {
          Rule rule = (Rule) (ruleIterator.next());
          String service = rule.getId().toString();
          Element serviceElement = doc.createElement("service");
          serviceElement.setAttribute("name", service);
          roleElement.appendChild(serviceElement);
          Target target = rule.getTarget();
          List actions = target.getActions();

          Iterator actionIterator = actions.iterator();
          while (actionIterator.hasNext()) {
            List action = (List) (actionIterator.next());
            Iterator matchIterator = action.iterator();
            while (matchIterator.hasNext()) {
              TargetMatch tm = (TargetMatch) (matchIterator.next());
              StringAttribute sa = (StringAttribute) (tm.
                  getMatchValue());
              String method = sa.getValue();
              Element methodElement = doc.createElement("method");
              methodElement.setAttribute("name", method);
              serviceElement.appendChild(methodElement);
            }
          }
        }
        return roleElement;
      }
    }
    catch (Exception e) {
      throw AxisFault.makeFault(e);
    }
  }

  public static void main(String[] args) throws Exception {
    Node n = PermissionManager.listAssignedPermissions("administrator");
    System.out.println(org.apache.axis.utils.XMLUtils.ElementToString( (
        Element) n));
  }

}
