package net.starmiddleware.starws2.security.rbac;

import com.sun.xacml.*;
import com.sun.xacml.attr.*;
import com.sun.xacml.combine.*;
import com.sun.xacml.cond.*;
import com.sun.xacml.ctx.*;
import com.sun.xacml.finder.*;
import com.sun.xacml.finder.impl.*;

import java.net.URI;
import java.net.URISyntaxException;

import java.util.HashSet;
import java.util.Set;

/**
 * <p>Title: </p>
 * <p>Description:
 * This is an example program that shows how to build and generate an XACML
 * Request. This is a major part of what a PEP does. An equivalent Request to
 * that generated here is found in the request directory as generated.xml.
 * The generated policy can be used with this request.
 * </p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: </p>
 * @author zm
 * @version 1.0
 */

public class RequestProducer {

    /**
     * Sets up the Subject section of the request. This Request only has
     * one Subject section, and it uses the default category. To create a
     * Subject with a different category, you simply specify the category
     * when you construct the Subject object.
     *
     * @return a Set of Subject instances for inclusion in a Request
     *
     * @throws URISyntaxException if there is a problem with a URI
     */
    public static Set setupSubjects(String role) throws URISyntaxException {
	HashSet attributes = new HashSet();

	// setup the id and value for the requesting subject
	URI subjectId = new URI("role");

	StringAttribute value = new StringAttribute(role);

	// create the subject section with two attributes, the first with
	// the subject's identity...
	attributes.add(new Attribute(subjectId, null, null, value));

	// bundle the attributes in a Subject with the default category
	HashSet subjects = new HashSet();
	subjects.add(new Subject(attributes));

	return subjects;
    }

    /**
     * Creates a Resource specifying the resource-id, a required attribute.
     *
     * @return a Set of Attributes for inclusion in a Request
     *
     * @throws URISyntaxException if there is a problem with a URI
     */

    public static Set setupResource(String service) throws URISyntaxException {
	HashSet resource = new HashSet();

	// the resource being requested
	StringAttribute value = new StringAttribute(service);

	// create the resource using a standard, required identifier for
	// the resource being requested
	resource.add(new Attribute(new URI(EvaluationCtx.RESOURCE_ID), null, null,
				   value));

	return resource;
    }

    /**
     * Creates an Action specifying the action-id, an optional attribute.
     *
     * @return a Set of Attributes for inclusion in a Request
     *
     * @throws URISyntaxException if there is a problem with a URI
     */

    public static Set setupAction(String method) throws URISyntaxException {
	HashSet action = new HashSet();

	// this is a standard URI that can optionally be used to specify
	// the action being requested
	URI actionId = new URI("method");

	// create the action
	action.add(new Attribute(actionId, null, null,
				 new StringAttribute(method)));

	return action;
    }

    /**
     * Command-line interface that creates a new Request by invoking the
     * static methods in this class. The Request has no Environment section.
     */
    public static void main(String[] args) throws Exception {

	String role = "employee";
	String service = "test";
	// "urn:xmltoday-delayed-quotes";
	String method = "renewQuote";
	//"getQuote";

	// create the new Request...note that the Environment must be specified
	// using a valid Set, even if that Set is empty
	RequestCtx request = new RequestCtx(setupSubjects(role),
					    setupResource(service),
					    setupAction(method), new HashSet());

	// encode the Request and print it to standard out
	request.encode(System.out, new Indenter());

	//test!!
	java.io.File file = new java.io.File("test.xml");
	java.io.FileOutputStream output = new java.io.FileOutputStream(file);
	request.encode(output);
	output.close();

    }

    public static RequestCtx makeRequest(String role, String service, String method)throws Exception{
	RequestCtx request = new RequestCtx(setupSubjects(role),
					    setupResource(service),
					    setupAction(method), new HashSet());
        return request;
    }

}
