package net.starmiddleware.starws2.security.rbac;

import org.apache.axis.Constants;
import org.apache.axis.Message;
import org.apache.axis.MessageContext;
import org.apache.axis.client.AxisClient;
import org.apache.axis.configuration.NullProvider;
import org.apache.axis.encoding.DeserializationContextImpl;
import org.apache.axis.encoding.SerializationContext;
import org.apache.axis.encoding.SerializationContextImpl;
import org.apache.axis.message.SOAPEnvelope;
import org.apache.axis.message.SOAPHeaderElement;
import org.apache.axis.utils.Mapping;
import org.apache.axis.utils.Messages;
import org.apache.axis.utils.XMLUtils;
import org.apache.xml.security.c14n.Canonicalizer;
import org.apache.xml.security.signature.XMLSignature;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.InputSource;
import org.apache.axis.handlers.BasicHandler;

import java.io.FileInputStream;
import java.io.Reader;
import java.io.StringReader;
import java.io.StringWriter;
import java.security.KeyStore;
import java.security.PrivateKey;
import java.security.cert.X509Certificate;

/**
 * <p>Title: </p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2002</p>
 * <p>Company: </p>
 * @author zm
 * @version 1.0
 */



public class SignedSOAPEnvelope
    extends SOAPEnvelope {

  //namespaceURI and alias
  static String SOAPSECNS = "http://schemas.xmlsoap.org/soap/security/2000-12";
  static String SOAPSECprefix = "SOAP-SEC";

  //necessary parameters in .wsdd file
  static String KEYSTORE_TYPE = "keystoreType";
  static String KEYSTORE_FILE = "keystoreFile";
  static String KEYSTORE_PASS = "keystorePass";
  static String PRIVATE_KEY_ALIAS = "privateKeyAlias";
  static String PRIVATE_KEY_PASS = "privateKeyPass";
  static String CERTIFICATE_ALIAS = "certificateAlias";
  static String MESSAGE_TYPE = "messageType";

  private MessageContext msgContext;
  private BasicHandler handler;

  static {
    org.apache.xml.security.Init.init();
  }

  public SignedSOAPEnvelope(MessageContext msgContext, SOAPEnvelope env,
                            String baseURI, BasicHandler handler) {
    this.msgContext = msgContext;
    this.handler = handler;
    init(env, baseURI);
  }

  public SignedSOAPEnvelope(SOAPEnvelope env, String baseURI) {
    init(env, baseURI);
  }

  private void init(SOAPEnvelope env, String baseURI) {
    try {
      String relativeRootPath = RBACUtil.getSystemRootPath();
      //(String) handler.getOption("relativeRootPath");
      //System.getProperty("starws.home")

      String keystoreType = (String) handler.getOption(KEYSTORE_TYPE);
      String keystoreFile = relativeRootPath +
          (String) handler.getOption(KEYSTORE_FILE);
      String keystorePass = (String) handler.getOption(KEYSTORE_PASS);
      String privateKeyAlias = (String) handler.getOption(PRIVATE_KEY_ALIAS);
      String privateKeyPass = (String) handler.getOption(PRIVATE_KEY_PASS);
      String certificateAlias = (String) handler.getOption(CERTIFICATE_ALIAS);
      String type = (String) handler.getOption(MESSAGE_TYPE);

      System.out.println("Beginning signing...");

      //SOAPEnvelopeһЩ
      env.addMapping(new Mapping(SOAPSECNS, SOAPSECprefix));
      env.addAttribute(Constants.URI_SOAP11_ENV, "actor", "some-uri");
      env.addAttribute(Constants.URI_SOAP11_ENV, "mustUnderstand", "1");

      SOAPHeaderElement header =
          new SOAPHeaderElement(XMLUtils.StringToElement(SOAPSECNS,
          "Signature",
          ""));
      env.addHeader(header);

      Document doc = getSOAPEnvelopeAsDocument(env, msgContext);

      KeyStore ks = KeyStore.getInstance(keystoreType);
      FileInputStream fis = new FileInputStream(keystoreFile);

      ks.load(fis, keystorePass.toCharArray());

      PrivateKey privateKey = (PrivateKey) ks.getKey(privateKeyAlias,
          privateKeyPass.toCharArray());

      Element soapHeaderElement = (Element) ( (Element) doc.getFirstChild()).
          getElementsByTagNameNS("*", "Header").item(0);
      Element soapSignatureElement = (Element) soapHeaderElement.
          getElementsByTagNameNS("*", "Signature").item(0);

      XMLSignature sig = new XMLSignature(doc, baseURI,
                                          XMLSignature.ALGO_ID_SIGNATURE_DSA);

      soapSignatureElement.appendChild(sig.getElement());
      sig.addDocument("#Body");

      X509Certificate cert =
          (X509Certificate) ks.getCertificate(certificateAlias);

      sig.addKeyInfo(cert);
      sig.addKeyInfo(cert.getPublicKey());
      sig.sign(privateKey);

      Canonicalizer c14n = Canonicalizer.getInstance(Canonicalizer.
          ALGO_ID_C14N_WITH_COMMENTS);
      byte[] canonicalMessage = c14n.canonicalizeSubtree(doc);

      InputSource is = new InputSource(new java.io.ByteArrayInputStream(
          canonicalMessage));
      DeserializationContextImpl dser = null;
      if (msgContext == null) {
        AxisClient tmpEngine = new AxisClient(new NullProvider());
        msgContext = new MessageContext(tmpEngine);
      }
      dser = new DeserializationContextImpl(is, msgContext,
                                            type, this);

      dser.parse();
      System.out.println(" signing complete.");
    }
    catch (Exception e) {
      e.printStackTrace();
      throw new RuntimeException(e.toString());
    }
  }

  private Document getSOAPEnvelopeAsDocument(SOAPEnvelope env,
                                             MessageContext msgContext) throws
      Exception {
    StringWriter writer = new StringWriter();
    SerializationContext serializeContext = new SerializationContextImpl(writer,
        msgContext);
    env.output(serializeContext);
    writer.close();

    Reader reader = new StringReader(writer.getBuffer().toString());
    Document doc = XMLUtils.newDocument(new InputSource(reader));
    if (doc == null)
      throw new Exception(
          Messages.getMessage("noDoc00", writer.getBuffer().toString()));
    return doc;
  }
}
