package net.starmiddleware.starws2.security.rbac;

import org.apache.axis.AxisFault;
import org.apache.axis.Handler;
import org.apache.axis.Message;
import org.apache.axis.MessageContext;
import org.apache.axis.handlers.BasicHandler;
import org.apache.axis.components.logger.LogFactory;
import org.apache.commons.logging.Log;
import org.apache.axis.utils.Messages;

import org.apache.xml.security.signature.XMLSignature;
import org.apache.xml.security.utils.Constants;
import org.apache.xpath.CachedXPathAPI;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.io.FileWriter;
import java.io.PrintWriter;

/**
 * <p>Title: </p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2002</p>
 * <p>Company: </p>
 * @author zm
 * @version 1.0
 */


public class ValidateHandler
    extends BasicHandler {

  //xmlsec.jar needs initialization
  static {
    org.apache.xml.security.Init.init();
  }

  //implement the BasicHandler's abstract method
  public void invoke(MessageContext msgContext) throws AxisFault {
    try {

      Message inMsg = msgContext.getRequestMessage();
      Message outMsg = msgContext.getResponseMessage();

      // verify signed message
      Document doc = msgContext.getCurrentMessage().getSOAPEnvelope().
          getAsDocument();
      String BaseURI = "http://xml-security";
      CachedXPathAPI xpathAPI = new CachedXPathAPI();

      Element nsctx = doc.createElement("nsctx");
      nsctx.setAttribute("xmlns:ds", Constants.SignatureSpecNS);

      Element signatureElem = (Element) xpathAPI.selectSingleNode(doc,
          "//ds:Signature", nsctx);

      // check to make sure that the document claims to have been signed
      if (signatureElem == null) {
        System.out.println("The document is not signed");
        throw new AxisFault("ValidateHandler", "not signed", null, null);
      }
      XMLSignature sig = new XMLSignature(signatureElem, BaseURI);

      boolean verify = sig.checkSignatureValue(sig.getKeyInfo().getPublicKey());
      // System.out.println(" verification complete.!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");

      // System.out.println("The signature is" + (verify? " " : " not ") + "valid");
      if (!verify)
        throw new AxisFault("ValidateHandler", "verify not valid", null, null);

      java.security.cert.X509Certificate cert = sig.getKeyInfo().
          getX509Certificate();

      String relativeRootPath = RBACUtil.getSystemRootPath();
          //(String) getOption("relativeRootPath");
      String CACertificateFile = relativeRootPath + (String)getOption("CACertificateFile");

      if (!CheckCertSign.verify(cert, CACertificateFile)) {
        throw new AxisFault("ValidateHandler", "the certificate is not valid", null, null);
      }

      String usernameInCert = cert.getSubjectX500Principal().getName();
      msgContext.setUsername(usernameInCert);
      // System.out.println(username);
      String username = msgContext.getUsername();
      //if( usernameInCert.startsWith("CN=")){
      //if( !usernameInCert.substring(3).startsWith(username)){
      if (username != null) {
        if (!username.equals(usernameInCert))
          throw new AxisFault("ValidateHandler",
              "the username in the certificate does not match the username logged in",
              null, null);
      }
      //}else{
      //  throw new AxisFault("ValidateHandler", "username in the certificate is wrong in formation", null, null);
      // }

    }
    catch (Exception e) {
      throw AxisFault.makeFault(e);
    }

  }
}
