package net.starmiddleware.starws2.transport.http.admin.deploy;

/**
 * <p>Title: CCM component Service Description</p>
 * <p>Description:  descripe CCM component information when deploying facet services to web services  </p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: StarMiddleware.net</p>
 * @author Wang Hongjun              flymanmail@163.com
 * @version 1.0
 */

import java.io.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;

import org.apache.axis.*;
import org.apache.axis.utils.*;
import org.omg.CORBA.*;
import org.omg.CORBA.InterfaceDefPackage.*;
import STARCCM.client.Components.*;
import net.starmiddleware.starws2.transport.http.admin.*;

public class CCMServiceDeploy {

  //variables used in deployment
  //by jeff won
  protected ORB orb = null;
  protected Repository repository = null;
  protected HomeFinder homeFinder = null;


  protected String name = "ccm-HelloWorld";
  protected String handlerClass =
      "net.starmiddleware.starws2.providers.ccm.CCMProvider";
  protected String repositoryURL = "relfile:/irior";
  protected String homeFinderURL = "corbaloc::172.26.13.66:9992/HomeFinder";
  protected String componentName = "HelloWorld";
  protected String interfaceName = "HelloWorld::Person";
  protected String homeInterfaceName = "HelloWorld::PersonHome";
  protected String facetInterfaceNames = "HelloWorld::Singer";
  protected String ORBClass = "net.nis.CORBA.ORB";
  protected String ORBSingletonClass = "net.nis.CORBA.ORBSingleton";
  protected String allowedMethods = "";
  protected String scope = "session";

  protected static String[] localMethods = {
      "get_component_def", "get_ccm_home", "get_primary_key"
      , "configuration_complete", "remove", "get_all_ports"
      , "provide_facet", "get_all_facets", "get_named_facets"
      , "same_component", "connect", "disconnect"
      , "get_connections", "get_all_receptacles", "get_named_receptacles"
      , "get_consumer", "subscribe", "unsubscribe"
      , "connect_consumer", "disconnect_consumer", "get_all_consumers"
      , "get_named_consumers", "get_all_emitters", "get_named_emitters"
      , "get_all_publishers", "get_named_publishers"
  };


  public String toString() {
    StringBuffer result = new StringBuffer();

    //<service...>
    result.append("<service ");
    if (name != null) {
      result.append("name=\"" + name + "\" ");
      result.append("provider=\"" + "Handler" + "\"");
    }
    result.append(">");

    //<handlerClass...>
    if (handlerClass != null) {
      result.append("<parameter name=\"handlerClass\" value=\"" + handlerClass +
                    "\"/>");
    }

    //<homeFinderURL...>
    if (homeFinderURL != null) {
      result.append("<parameter name=\"homeFinderURL\" value=\"" +
                    homeFinderURL + "\"/>");
    }

    //<repositoryURL...>
    if (repositoryURL != null) {
      result.append("<parameter name=\"repositoryURL\" value=\"" +
                    repositoryURL + "\"/>");
    }

    //<componentName...>
    if (componentName != null) {
      result.append("<parameter name=\"componentName\" value=\"" +
                    componentName + "\"/>");
    }

    //<interfaceName...>
    if (interfaceName != null) {
      result.append("<parameter name=\"interfaceName\" value=\"" +
                    interfaceName + "\"/>");
    }

    //<homeInterfaceName...>
    if (homeInterfaceName != null) {
      result.append("<parameter name=\"homeInterfaceName\" value=\"" +
                    homeInterfaceName + "\"/>");
    }

    //<facetInterfaceNames...>
    if (interfaceName != null) {
      result.append("<parameter name=\"facetInterfaceNames\" value=\"" +
                    facetInterfaceNames + "\"/>");
    }

    //<ORBClass...>
    if (ORBClass != null) {
      result.append("<parameter name=\"ORBClass\" value=\"" + ORBClass + "\"/>");
    }

    //<ORBSingletonClass...>
    if (ORBSingletonClass != null) {
      result.append("<parameter name=\"ORBSingletonClass\" value=\"" +
                    ORBSingletonClass + "\"/>");
    }

    //<allowedMethods...>
    if (!allowedMethods.equals("")) {
      result.append("<parameter name=\"allowedMethods\" value=\"" +
                    allowedMethods +
                    "\"/>");
    }
    else {
      result.append("<parameter name=\"allowedMethods\" value=\"*\"/>");
    }

    //<scope...>
    if (scope != null) {
      result.append("<parameter name=\"scope\" value=\"" + scope + "\"/>");
    }

    result.append("</service>");
    return result.toString();

  }

  protected ORB initORB(String orbcls, String orbsingletcls) throws
      AxisFault {
    ORB orb = null;

    java.util.Properties props = System.getProperties();
    props.put("org.omg.CORBA.ORBClass", orbcls);
    props.put("org.omg.CORBA.ORBSingletonClass", orbsingletcls);

    //
    //get ORB Obj
    //
    try {
      String[] args = new String[1];
      args[0] = "";
      orb = (ORB) ORB.init(args, props);

    }
    catch (Exception e) {
      throw new AxisFault(Messages.getMessage("initORBError"));
    }

    return (ORB) orb;
  }


  protected ArrayList getAllDeployMethods(HttpServletRequest request,
                                          HttpServletResponse response,
                                          ArrayList interfaceNames) throws
      AxisFault {

    ArrayList interfaceMethods = new ArrayList();

    for (int i = 0; i < interfaceNames.size(); i++) {
      String interfaceName = (String) interfaceNames.get(i);
      Contained ifContained = repository.lookup(interfaceName);
      assertNotNull(ifContained, "noInterfaceDefined");
      InterfaceDef ifDef = InterfaceDefHelper.narrow(ifContained);
      FullInterfaceDescription ifFullDesc = ifDef.describe_interface();
      OperationDescription[] opDescs = ifFullDesc.operations;
      ArrayList methods = new ArrayList();
      outer:
          for (int j = 0; j < opDescs.length; j++) {
        String method = opDescs[j].name;
        if (i == 0) {
          for (int k = 0; k < localMethods.length; k++) {
            if (method.equals(localMethods[k])) {
              continue outer;
            }
          }
        }
        methods.add(method);
      }
      interfaceMethods.add(methods);
    }
    return interfaceMethods;

  }

  public void addNewCCMService(HttpServletRequest request,
                               HttpServletResponse response) throws IOException,
      ServletException {
    request.setAttribute("name", name);
    request.setAttribute("repositoryURL", repositoryURL);
    request.setAttribute("homeFinderURL", homeFinderURL);
    request.setAttribute("componentName", componentName);
    request.setAttribute("interfaceName", interfaceName);
    request.setAttribute("homeInterfaceName", homeInterfaceName);
    request.setAttribute("facetInterfaceNames", facetInterfaceNames);
    request.setAttribute("ORBClass", ORBClass);
    request.setAttribute("ORBSingletonClass", ORBSingletonClass);
    request.setAttribute("scope", scope);
    request.getRequestDispatcher(
        "/admin/CCMService/addNewServiceBasicInfo.jsp").forward(request,
        response);
  }

  public void addNewCCMServiceBasicInfo(HttpServletRequest request,
                                        HttpServletResponse response) throws
      AxisFault, IOException, ServletException {
    //
    //get parameters from request
    //
    //service name
    name = (String) request.getParameter("name");
    if (name == null || name.length() < 1) {
      throw new AxisFault("no service name");
    }
    //repositoryURL
    repositoryURL = request.getParameter("repositoryURL");
    if (repositoryURL == null || repositoryURL.length() < 1) {
      throw new AxisFault("no repositoryURL");
    }
    //homeFinderURL
    homeFinderURL = request.getParameter("homeFinderURL");
    if (homeFinderURL == null || homeFinderURL.length() < 1) {
      throw new AxisFault("no homeFinderURL");
    }
    //componentName
    componentName = request.getParameter("componentName");
    if (componentName == null || componentName.length() < 1) {
      throw new AxisFault("no componentName");
    }
    //interfaceName
    interfaceName = request.getParameter("interfaceName");
    if (interfaceName == null || interfaceName.length() < 1) {
      throw new AxisFault("no interfaceName");
    }
    //homeInterfaceName
    homeInterfaceName = request.getParameter("homeInterfaceName");
    if (homeInterfaceName == null) {
      throw new AxisFault("no homeInterfaceName");
    }
    //facetInterfaceNames
    facetInterfaceNames = request.getParameter("facetInterfaceNames");

    //allowedMethods
    allowedMethods = request.getParameter("allowedMethods");
    if (allowedMethods == null) {
      allowedMethods = "";
    }

    //scope
    scope = request.getParameter("scope");
    if (scope == null || scope.length() < 1) {
      scope = "session";
    }
    //ORBClass
    ORBClass = request.getParameter("ORBClass");
    if (ORBClass == null) {
      throw new AxisFault("no ORBClass");
    }
    //ORBSingletonClass
    ORBSingletonClass = request.getParameter("ORBSingletonClass");
    if (ORBSingletonClass == null) {
      throw new AxisFault("no ORBSingletonClass");
    }

    //
    //init data fields
    //
    //init ORB Obj
    if (orb == null) {
      orb = initORB(ORBClass, ORBSingletonClass);
    }

    //init Repository Obj
    if (repository == null) {
      repository = initRepository(repositoryURL);
    }

    //init HomeFinder Obj
    if (homeFinder == null) {
      homeFinder = initHomeFinder(homeFinderURL);
    }

    //
    //get interfaceNames
    //
    ArrayList interfaceNames = new ArrayList();
    interfaceNames.add(interfaceName);

    StringTokenizer st = new StringTokenizer(facetInterfaceNames, " ");
    while (st.hasMoreTokens()) {
      String interfaceName = st.nextToken();
      interfaceNames.add(interfaceName);
    }

    //
    //get interfaceMethods
    //
    ArrayList interfaceMethods = getAllDeployMethods(request, response,
        interfaceNames);
    request.setAttribute("interfaceNames", interfaceNames);
    request.setAttribute("interfaceMethods", interfaceMethods);

  }

  public void addNewCCMServiceMethodsInfo(HttpServletRequest request,
                                          HttpServletResponse response) {

    //String[] methods = (String[]) request.getAttribute("methods");
    String[] methods = request.getParameterValues("methods");
    if (methods == null) {
      allowedMethods = "*";
    }
    else {

      allowedMethods = "";
      for (int i = 0; i < methods.length; i++) {
        allowedMethods += methods[i];
        allowedMethods += " ";
      }
    }

  }

  protected void assertNotNull(java.lang.Object obj, String str) throws
      AxisFault {
    if (obj == null) {
      throw new AxisFault(Messages.getMessage(str));
    }
  }

  protected Repository initRepository(String repositoryURL) throws AxisFault {

    assertNotNull(orb, "noORB");
    if ( (repositoryURL == null) || repositoryURL.length() < 1) {
      throw new AxisFault(Messages.getMessage(
          "initRepositoryError"));
    }

    //
    //get Rpository Obj
    //
    Repository repository = null;
    try {
      org.omg.CORBA.Object obj = orb.string_to_object(repositoryURL);
      repository = RepositoryHelper.narrow(obj);
    }
    catch (Exception e) {
      throw new AxisFault(Messages.getMessage(
          "initRepositoryError"));
    }

    return repository;
  }

  protected HomeFinder initHomeFinder(String homeFinderURL) throws AxisFault {

    assertNotNull(orb, "noORB");
    if ( (homeFinderURL == null) || homeFinderURL.length() < 1) {
      throw new AxisFault(Messages.getMessage(
          "initHomeFinderError"));
    }

    //
    //get HomeFinder Obj
    //
    HomeFinder homeFinder = null;
    try {
      org.omg.CORBA.Object obj = orb.string_to_object(homeFinderURL);
      homeFinder = HomeFinderHelper.narrow(obj);
    }
    catch (Exception e) {
      throw new AxisFault(Messages.getMessage(
          "initHomeFinderError"));
    }

    return homeFinder;
  }

  public void ccmModifyServiceSuccess(HttpServletRequest request,
                                      HttpServletResponse response,
                                      AdminService admin) throws
      ServletException, IOException {

    addNewCCMServiceBasicInfo(request, response);

    String body = "<deployment xmlns=\"http://xml.apache.org/axis/wsdd/\"  xmlns:java=\"http://xml.apache.org/axis/wsdd/providers/java\">" +
        this.toString() + "</deployment>";
    MessageContext msgContext = admin.invokeAdmin(request, response, body);
    if (msgContext != null) {
      request.setAttribute("info", "Service is modified successfully.");
      request.getRequestDispatcher("../admin/info.jsp").forward(request,
          response);
    }
  }
}
