/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2001-2003 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Axis" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

package net.starmiddleware.starws2.util.CORBA;
/**
 * <p>Title: convert</p>
 * <p>Description: convert data type from CORBA's any type to CORBA type,
 * or from CORBA type to CORBA's any type</p>
 * <p>group:starMiddleware </p>
 * @author shuichao,(super_shuichao@163.com)
 * @author huJianQiang
 * @version 1.0
 */
import org.omg.CORBA.Any;
import org.omg.CORBA.TypeCode;
import org.omg.CORBA.TypedefDef;
import org.omg.CORBA.TypedefDefHelper;
import org.omg.DynamicAny.*;
import org.omg.CORBA.Object;
import org.omg.CORBA.ORB;
import org.omg.CORBA.InterfaceDef;
import org.omg.CORBA.TCKind;
import java.util.*;
import net.starmiddleware.starws2.encoding.CORBA.*;
import net.starmiddleware.starws2.encoding.CORBA.ComplexType;
import javax.xml.namespace.QName;

public class Convert {

  private QName defaultType=null;

  public Convert() {
  }
  /**
   *
   * @param _orb:ORB
   * @param ctypecode:
   * @param complex:
   * @return
   */
  public  Any simple2Any(ORB _orb,int typeCodeValue,java.lang.Object param)
      throws  IllegalArgumentException{

    Any any= _orb.create_any();

    switch (typeCodeValue) {
      case org.omg.CORBA.TCKind._tk_char:
       {
      try {
        if(param==null){
          any.insert_char('0');
        }
        if (param instanceof java.lang.Character) {
          any.insert_char(((java.lang.Character )param).charValue());
        }
        else if((param instanceof String)&&(((String)param).length()==1)){
           any.insert_char(((String )param).charAt(0));
        }
        else
          throw new IllegalArgumentException(new Exception("con not convert " +
              param.toString() + " to corba_char"));
        } //end of if
      catch (Exception e) {
        throw new IllegalArgumentException(new Exception(
            "Error in char to any"));
      }
    } //end of case
    break;
    case org.omg.CORBA.TCKind._tk_wchar:
       {
      try {
        if (param instanceof java.lang.Character) {
          any.insert_wchar(((java.lang.Character )param).charValue());
        }
        else if((param instanceof String)&&(((String)param).length()==1)){
           any.insert_wchar(((String )param).charAt(0));
        }

        else
          throw new IllegalArgumentException(new Exception("con not convert " +
              param.toString() + " to corba_wchar"));
        } //end of if
      catch (Exception e) {
        throw new IllegalArgumentException(new Exception(
            "Error in wchar to any"));
      }
    } //end of case
    break;

      case org.omg.CORBA.TCKind._tk_octet:
       {
      try {
        if (param instanceof org.apache.axis.types.UnsignedByte) {
          any.insert_octet(((org.apache.axis.types.UnsignedByte)param).byteValue());
        }
        else
          throw new IllegalArgumentException(new Exception("con not convert " +
              param.toString() + " to corba_octet"));
        } //end of if
      catch (Exception e) {
        throw new IllegalArgumentException(new Exception(
            "Error in octet to any"));
      }
    } //end of case
    break;

        case org.omg.CORBA.TCKind._tk_long:
        {
       try {
         if (param instanceof java.lang.Integer) {
           any.insert_long(((Integer)param).intValue());
         }
         else if(param instanceof org.apache.axis.types.UnsignedInt){
           any.insert_long(((org.apache.axis.types.UnsignedInt)param).intValue());
         }
         else
           throw new IllegalArgumentException(new Exception("con not convert " +
               param.toString() + " to corba_long"));
         } //end of if
       catch (Exception e) {
         throw new IllegalArgumentException(new Exception(
             "Error in integer to any"));
       }
     } //end of case
     break;

        case org.omg.CORBA.TCKind._tk_ulong:
        {
          try {
            if (param instanceof java.lang.Integer) {
              any.insert_ulong(((Integer)param).intValue());
            }
            else if(param instanceof org.apache.axis.types.UnsignedInt){
              any.insert_ulong(((org.apache.axis.types.UnsignedInt)param).intValue());
            }
            else
              throw new IllegalArgumentException(new Exception("con not convert " +
                  param.toString() + " to corba_ulong"));
            } //end of if
          catch (Exception e) {
            throw new IllegalArgumentException(new Exception(
                "Error in ulong to any"));
          }
        } //end of case
        break;



        case org.omg.CORBA.TCKind._tk_float: {

          try {
            if (param instanceof java.lang.Float) {
              any.insert_float(((Float)param).floatValue());
            }
            else {
              throw new IllegalArgumentException(new Exception("con not convert " +
                  param.toString() + " to corba_float"));
            }

          }
          catch (Exception e) {
            throw new IllegalArgumentException(new Exception(
                "Error in float to any"));
          }
        } //end of case org.omg.CORBA.TCKind._tk_float :
        break;

        case org.omg.CORBA.TCKind._tk_short:
        {
        try {
          if (param instanceof java.lang.Short) {
            Short x = (Short) param;
            any.insert_short(x.shortValue());
          }
          else if(param instanceof org.apache.axis.types.UnsignedShort){
            any.insert_short(((org.apache.axis.types.UnsignedShort)param).shortValue());
          }
          else
            throw new IllegalArgumentException(new Exception("con not convert " +
                param.toString() + " to corba_short"));
          }
        catch (Exception e) {
          throw new IllegalArgumentException(new Exception(
              "Error in short to any"));
        }
        }
        break;

        case org.omg.CORBA.TCKind._tk_ushort:
        {
          try {
            if (param instanceof java.lang.Short) {
              Short x = (Short) param;
              any.insert_ushort(x.shortValue());
            }
            else if(param instanceof org.apache.axis.types.UnsignedShort){
              any.insert_ushort(((org.apache.axis.types.UnsignedShort)param).shortValue());
            }
            else
              throw new IllegalArgumentException(new Exception("con not convert " +
                  param.toString() + " to corba_ushort"));
            }
          catch (Exception e) {
            throw new IllegalArgumentException(new Exception(
                "Error in ushort to any"));
          }
          ;
        } //end of case
        break;



        case org.omg.CORBA.TCKind._tk_longlong:
        {
        try {
          if (param instanceof java.lang.Long) {
            Long x = (Long) param ;
            any.insert_longlong(x.longValue());
          }
          else if(param instanceof org.apache.axis.types.UnsignedLong){
            any.insert_ulonglong(((org.apache.axis.types.UnsignedLong)param).shortValue());
          }

          else {
            throw new IllegalArgumentException(new Exception("con not convert " +
                param.toString() + " to corba_luonglong"));
          }

        }
        catch (Exception e) {
          throw new IllegalArgumentException(new Exception(
              "Error in ulonglong to any"));
        }
        ;
      } //end of case
      break;


        case org.omg.CORBA.TCKind._tk_ulonglong:
        {
          try {
            if (param instanceof java.lang.Long) {
              Long x = (Long) param ;
              any.insert_longlong(x.longValue());
            }
            else if(param instanceof org.apache.axis.types.UnsignedLong){
              any.insert_longlong(((org.apache.axis.types.UnsignedLong)param).shortValue());
            }

            else {
              throw new IllegalArgumentException(new Exception("con not convert " +
                  param.toString() + " to corba_longlong"));
            }

          }
          catch (Exception e) {
            throw new IllegalArgumentException(new Exception(
                "Error in longlong to any"));
          }
          ;
        } //end of case
        break;



        case org.omg.CORBA.TCKind._tk_string:
        {
        try{
          if (param instanceof java.lang.String) {
            any.insert_string( (String)param);
          }
          else {
            throw new IllegalArgumentException(new Exception("con not convert " +
                param.toString() + " to corba_String"));
          }
        }catch (Exception e) {
          throw new IllegalArgumentException(new Exception(
                "Error in String to any"));
        }

        } //end of if
          break;

        case org.omg.CORBA.TCKind._tk_wstring:
        {
        try{
          if (param instanceof java.lang.String) {
            any.insert_wstring( (String)param);
          }
          else {
            throw new IllegalArgumentException(new Exception("con not convert " +
                param.toString() + " to corba_wString"));
          }
        }catch (Exception e) {
          throw new IllegalArgumentException(new Exception(
                "Error in wString to any"));
        }

        } //end of if
          break;


        case org.omg.CORBA.TCKind._tk_double: {
          try {
//                              System.out.print(param.getType());
            if (param instanceof java.lang.Double) {
              Double x = (Double) param ;
              any.insert_double(x.doubleValue());

            }
          else{
            throw new IllegalArgumentException(new Exception("con not convert "+param.toString()+" to corba_double"));
           }

          }
          catch (Exception e) {
            throw new IllegalArgumentException(new Exception(
                "Error in double to any"));
          }

        } //end of case
        break;

        case org.omg.CORBA.TCKind._tk_boolean: {
          try {
//                              System.out.print(param.getType());
            if (param instanceof java.lang.Boolean) {
              Boolean x = (Boolean) param ;
              any.insert_boolean(x.booleanValue());
            }
          else{
            throw new IllegalArgumentException(new Exception("con not convert "+param.toString()+" to corba_boolean"));
          }

          }
          catch (Exception e) {
            throw new IllegalArgumentException(new Exception(
                "Error in boolean to any"));
          }
          ;
        } //end of case
        break;

        default:
          throw new IllegalArgumentException(
              new Exception("This data type is not supported in corba type"));
      } //end of case
      return any;
    }
 /*************************************************************************************/


 /*************************************************************************************/

  public  Any complex2Any(ORB _orb, TypeCode ctypecode,java.lang.Object complex)
      throws  IllegalArgumentException{

      Any  result=_orb.create_any();
      int num=0;

      Object objFact=null;
      try{
            objFact=_orb.resolve_initial_references("DynAnyFactory") ;

          }catch(Exception e){
              System.out.println("error in resolve DynAnyFacotry");
              throw new IllegalArgumentException(e);
          };

      DynAnyFactory factory=DynAnyFactoryHelper.narrow(objFact);
      org.omg.DynamicAny.DynAny  da=null;
      try{
          da = factory.create_dyn_any_from_type_code(ctypecode);

         }catch (Exception e) {
            System.out.println(e.toString());
            throw new IllegalArgumentException(e);
         }

      switch (ctypecode.kind().value())
      {
        case org.omg.CORBA.TCKind._tk_any:
          org.omg.DynamicAny.DynAny dynAny = org.omg.DynamicAny.DynAnyHelper.narrow(da) ;
          da.destroy();
           try{
             if (complex instanceof java.util.ArrayList) {

             }
             else {
               throw new IllegalArgumentException(new Exception("con not convert " +
                   complex.toString() + " to corba_any"));
             }
           }catch (Exception e) {
             throw new IllegalArgumentException(new Exception(
                 "Error in String to any"));
           }

        case org.omg.CORBA.TCKind._tk_alias:
         try{
           //org.omg.DynamicAny.DynAny dynAlias = org.omg.DynamicAny.DynArrayHelper.narrow(da) ;
           da.destroy();

           TypeCode base=ctypecode.content_type();
           if(isSimpleType(base.kind().value())) simple2Any(_orb,base.kind().value(),complex);
           else{
                 Any t_any=complex2Any(_orb,base,complex);
                 //t_any.type(ctypecode);
                 //da.insert_any(t_any);
                 //result=da.to_any();
                 result=t_any;

               }

         }catch(org.omg.CORBA.TypeCodePackage.BadKind e){
           throw new IllegalArgumentException(e);
         }


          break;
      case org.omg.CORBA.TCKind._tk_enum:
        org.omg.DynamicAny.DynEnum dynEnum = org.omg.DynamicAny.DynEnumHelper.narrow(da) ;
          da.destroy();
           try{
               ComplexType value=(ComplexType)complex;
               dynEnum.set_as_string((String)value.getValue());
                result=dynEnum.to_any();
           }catch (Exception e) {
             throw new IllegalArgumentException(new Exception(
                 "Error in String to any"));
           }


           break;

        case org.omg.CORBA.TCKind._tk_array:

          org.omg.DynamicAny.DynArray dynArray = org.omg.DynamicAny.DynArrayHelper.narrow(da) ;
          da.destroy();

          try{
               num=dynArray.component_count() ;

            Any[] NVPair = new Any[num];
            for(int i=0;i<num;i++){
               java.lang.Object param=null;
               if(complex instanceof ComplexType){
                 ArrayList params=((ComplexType)complex).getChild();
                 param=params.get(i*2);
               }
               else if(complex instanceof org.apache.axis.encoding.ser.ArrayDeserializer.ArrayListExtension){
                  param=((org.apache.axis.encoding.ser.ArrayDeserializer.ArrayListExtension)complex).get(i);
               }
               else if(complex instanceof ArrayList){
                 param=((ArrayList)complex).get(i);
              }

              else param=java.lang.reflect.Array.get(complex,i);

               Any[] elements=dynArray.get_elements();
               if(isSimpleType(elements[i].type().kind().value())){
                 NVPair[i]=simple2Any(_orb,elements[i].type().kind().value(),param);
                }
               else   NVPair[i]=complex2Any(_orb,elements[i].type(),param);
            }

                 dynArray.set_elements(NVPair);
           }catch (Exception e){
              throw new IllegalArgumentException(e);
           };

            result=dynArray.to_any();
            break;
      case org.omg.CORBA.TCKind._tk_union:


             org.omg.DynamicAny.DynUnion dynUnion = org.omg.DynamicAny.DynUnionHelper.narrow(da) ;
             da.destroy();

             try{
               ComplexType value=(ComplexType)complex;
              ArrayList child=value.getChild();

              dynUnion.get_discriminator().insert_long(((Integer)child.get(0)).intValue());

               if(isSimpleType(dynUnion.member_kind().value())){
                  DynAny dyn=dynUnion.member();
                  Any temp=simple2Any(_orb,dynUnion.member_kind().value(),child.get(2));
                  dyn.from_any(temp);

               }else{
                  Any temp=complex2Any(_orb,dynUnion.member().type(),child.get(2));
                  dynUnion.insert_any(temp);
               }

              }catch (Exception e){
                 throw new IllegalArgumentException(e);
              };

               result=dynUnion.to_any();
               break;
      case org.omg.CORBA.TCKind._tk_fixed:
         org.omg.DynamicAny.DynFixed dynFixed = org.omg.DynamicAny.
              DynFixedHelper.narrow(da);
          da.destroy();
          try{
            ComplexType value=(ComplexType)complex;
            java.math.BigDecimal temp=new java.math.BigDecimal((String)value.getValue());
            String a=( (java.math.BigDecimal) temp).toString();
            dynFixed.set_value(a);
          }catch(Exception e){
            throw new IllegalArgumentException(new Exception("can not covert " +
                complex.toString() + " to array"));
          }

          result=dynFixed.to_any();

          break;

      case org.omg.CORBA.TCKind._tk_struct:


          org.omg.DynamicAny.DynStruct dynStruct = org.omg.DynamicAny.DynStructHelper.narrow(da) ;
          da.destroy();

          try{
            num = dynStruct.component_count();
            ComplexType value=(ComplexType)complex;
            org.omg.DynamicAny.NameValuePair[] NVPair = new org.omg.DynamicAny.
                NameValuePair[num];
            org.omg.DynamicAny.NameDynAnyPair[] elements = dynStruct.
                get_members_as_dyn_any();

            if (value.hasChilds()) {

              ArrayList params=value.getChild();
              for (int i = 0; i < num; i ++) {
                java.lang.Object param = params.get(i*2);

                if (isSimpleType(elements[i].value.type().kind().value())) {
                  Any simple = simple2Any(_orb,
                                          elements[i].value.type().kind().value(),
                                          param);
                  NVPair[i] = new org.omg.DynamicAny.NameValuePair(ctypecode.
                      member_name(i), simple);
                }
                else {
                  Any tComplex = complex2Any(_orb, elements[i].value.type(),
                                              param);
                  NVPair[i] = new org.omg.DynamicAny.NameValuePair(ctypecode.
                      member_name(i), tComplex);
                }
              }

              dynStruct.set_members(NVPair);
            }
           }
            catch (Exception e) {
              throw new IllegalArgumentException(e);
            }



            result=dynStruct.to_any();
            break;
        case org.omg.CORBA.TCKind._tk_sequence:
          ArrayList params =null;

          if (! (complex instanceof ComplexType)){
            if(complex.getClass().isArray()){
               params=new ArrayList();
               int length=java.lang.reflect.Array.getLength(complex);
               for(int i=0;i<length;i++){
                 params.add(java.lang.reflect.Array.get(complex,i));

                 params.add("");
               }
            }
            else throw new IllegalArgumentException(new Exception("can not convert sequence"));
          }
          else params=((ComplexType)complex ).getChild();

          org.omg.DynamicAny.DynSequence dynSeq = org.omg.DynamicAny.DynSequenceHelper.narrow(da) ;
          da.destroy();

              try{

                num=params.size();

                Any[] NVPair = new Any[num/2];

                java.lang.Object param = null;

                for (int i = 0; i < num/2; i++) {

                  param = params.get(i*2);
                  if (isSimpleType(ctypecode.content_type().kind().value())) {

                    NVPair[i] = simple2Any(_orb,
                                           ctypecode.content_type().kind().value(),
                                           param);
                  }
                  else {
                    NVPair[i] = complex2Any(_orb, ctypecode.content_type(), param);
                  }

                }
                dynSeq.set_elements(NVPair);
               }catch (Exception e){
                  throw new IllegalArgumentException(e);
               };

                result=dynSeq.to_any();
                break;

           }//end of case



            return result;

  };

  private  java.util.Stack recu=new java.util.Stack();
  private  boolean isSoapencoding=false;

  public  java.lang.Object any2Complex(ORB _orb,Any anyRes,TypeCode code)
   throws  IllegalArgumentException{
    if(recu.empty()){
      isSoapencoding=false;

    }
    recu.push(Boolean.FALSE);
    String name=null;
    java.lang.Object result=null;
    switch(code.kind().value()){
      case TCKind._tk_TypeCode:
             break;
      case TCKind._tk_any:
             break;
      case TCKind._tk_enum:
             try {
               org.omg.DynamicAny.DynEnum dynEnum=(org.omg.DynamicAny.DynEnum)
                   createDynAny_fromAny(_orb,code.kind().value(),anyRes);
               result=new ComplexType();
               ((ComplexType)result).setValue(dynEnum.get_as_string());
               ((ComplexType)result).setNamePostfix(dynEnum.type().name());

             }
             catch (Exception ex) {
               return null;
             }
           break;
      case TCKind._tk_alias:
          try{

            name=code.name();
            TypeCode t_Code = code.content_type();

            if(isSimpleType(code.kind().value())){
              result=any2Simple(anyRes,t_Code);
            }else{
              result=any2Complex(_orb,anyRes,t_Code);
              if(result instanceof ComplexType){

                   ((ComplexType)result).setNamePostfix(name);
              }

            }

          }catch(org.omg.CORBA.TypeCodePackage.BadKind e){
            throw new IllegalArgumentException(e);
          }




           break;
      case TCKind._tk_array:

           if(recu.size()>1){
             recu.pop();
             recu.push(Boolean.TRUE);
             isSoapencoding=true;
           }
           try{
             name=anyRes.type().name();

           }catch(Exception e){

           }

           result=getComplexConvertArray(_orb,anyRes,code,name);


           break;
      case TCKind._tk_union:
                try{
                org.omg.DynamicAny.DynUnion dynUnion =(org.omg.DynamicAny.DynUnion)createDynAny_fromAny(_orb,code.kind().value(),anyRes);
                 name=code.name();

                result=new ComplexType();
                ((ComplexType)result).setNamePostfix(name);

                ((ComplexType)result).add(new java.lang.Integer(dynUnion.get_discriminator().get_long()),"discriminator");

                java.lang.Object o=null;
                Any a=dynUnion.member().to_any();
                if(isSimpleType(dynUnion.member_kind().value())){
                  o=any2Simple(a,dynUnion.member_kind());
                }else{
                  o=any2Complex(_orb,a,dynUnion.member().get_typecode());
                }
                ((ComplexType)result).add(o,dynUnion.member_name());

                }catch(Exception e){
                  return null;
                }


                break;

    case TCKind._tk_sequence:
      if (recu.size() > 1) {
        recu.pop();
        recu.push(Boolean.TRUE);
        isSoapencoding=true;
      }


           try {
          org.omg.DynamicAny.DynSequence dynSeq = (org.omg.DynamicAny.DynSequence)
                 createDynAny_fromAny(_orb, code.kind().value(), anyRes);

             String childTypeName="";
             result = new net.starmiddleware.starws2.encoding.CORBA.ComplexType();
              if(dynSeq.get_length()>0){
                do {

                  DynAny current = dynSeq.current_component();
                  if (isSimpleType(current.type().kind().value())) {
                    ( (ComplexType) result).add(any2Simple(current.to_any(),
                        current.type()),"item");
                     childTypeName = getSimpleTypeName(current.type().kind().value());
                  }
                  else {
                    java.lang.Object temp = any2Complex(_orb, current.to_any(),
                        current.type());
                    if(temp instanceof ComplexType){
                      if(((ComplexType)temp).getNamePrefix().equals("")){
                        ((ComplexType)temp).setNamePrefix("item");
                      }
                      ( (ComplexType) result).add(temp,( (ComplexType)temp).getNamePrefix());
                    }
                    else ( (ComplexType) result).add(temp,"item");
                  }
                }
                while (dynSeq.next());
              }
              if(!childTypeName.equals("")){
                 ((ComplexType)result).setNamePostfix("ArrayOf"+childTypeName);
              }

           }
           catch (Exception e) {
             return null;
           }

           break;

         case TCKind._tk_struct:

           try {
             org.omg.DynamicAny.DynStruct dynStruct = (org.omg.DynamicAny.DynStruct)
                 createDynAny_fromAny(_orb, code.kind().value(), anyRes);

             int num = dynStruct.component_count();
             result = new net.starmiddleware.starws2.encoding.CORBA.ComplexType();
             name=dynStruct.type().name();

             ((ComplexType)result).setNamePostfix(name);
             do {
               DynAny current = dynStruct.current_component();
               if (isSimpleType(current.type().kind().value())) {
                 ( (ComplexType) result).add(any2Simple(current.to_any(),
                     current.type()),dynStruct.current_member_name());
               }
               else {
                 java.lang.Object temp = any2Complex(_orb, current.to_any(),
                                                     current.type());
                 if(temp instanceof ComplexType){
                   if (((ComplexType)temp).getNamePrefix().equals(""))
                     ((ComplexType)temp).setNamePrefix(dynStruct.current_member_name());

                 }
                 ( (ComplexType) result).add(temp,dynStruct.current_member_name());
               }
             }
             while (dynStruct.next());

           }
           catch (Exception e) {
             return null;
           }

           break;
         case TCKind._tk_fixed:
             try {
               org.omg.DynamicAny.DynFixed dynFixed = (org.omg.DynamicAny.DynFixed)
                   createDynAny_fromAny(_orb, code.kind().value(), anyRes);
               result=new ComplexType();
               ((ComplexType)result).setValue(dynFixed.get_value().toString());


             }
             catch (Exception e) {
               return null;
             }

             break;


    }
    if(result instanceof ComplexType){
       ( (ComplexType) result).setType(this.defaultType);
     }

       if((result instanceof ComplexType)&&(isSoapencoding)){

              String postfix=( (ComplexType) result).getNamePostfix();

              if(code.kind().value()==TCKind._tk_array){

              }
              else{
                ((ComplexType) result).setNamePostfix("_SE_" + postfix);

              }

    }
    if(recu.isEmpty()){
      isSoapencoding=false;
    }
    return result;
  }
  public  java.lang.Object any2Simple(Any anyRes,TypeCode code)
  throws  IllegalArgumentException{
    return any2Simple(anyRes,code.kind());
  }
  public  java.lang.Object any2Simple(Any anyRes,TCKind code)
  throws  IllegalArgumentException{
    java.lang.Object obj = null;
    switch (code.value()) {
      case TCKind._tk_octet: {
         obj = new org.apache.axis.types.UnsignedByte();
         ((org.apache.axis.types.UnsignedByte)obj).setValue(anyRes.extract_octet());
        break;
      }


      case TCKind._tk_string: {
        obj = (java.lang.String) anyRes.extract_string();
        break;
      }
      case TCKind._tk_wstring: {
        obj = (java.lang.String) anyRes.extract_wstring();
        break;
      }
      case TCKind._tk_char: {
        char x = anyRes.extract_char();
        obj = new _char(x);
        break;
      }
      case TCKind._tk_wchar: {
        char x = anyRes.extract_wchar();
        obj = new WChar(x);
        break;
      }

      case TCKind._tk_boolean: {
        boolean x = anyRes.extract_boolean();
        obj = new Boolean(x);
        break;
      }
      case TCKind._tk_long: {
        int x = anyRes.extract_long();
        obj = new Integer(x);
        break;
      }
      case TCKind._tk_ulong: {
        int x = anyRes.extract_ulong();
        obj = new  org.apache.axis.types.UnsignedInt(x);
        break;
      }
      case TCKind._tk_longlong: {

        long x = anyRes.extract_longlong();
        obj = new Long(x);
        break;
      }
      case TCKind._tk_ulonglong: {

        long x=0;
        try{
           x= anyRes.extract_ulonglong();
        }catch(Exception e){
          x= anyRes.extract_longlong();
        }
        obj = new  org.apache.axis.types.UnsignedLong(x);
        break;
      }
      case TCKind._tk_short: {
        short x = anyRes.extract_short();
        obj = new Short(x);
        break;
      }
      case TCKind._tk_ushort: {
        short x = anyRes.extract_ushort();
        obj = new org.apache.axis.types.UnsignedShort(x);
        break;
      }
      case TCKind._tk_float: {
        float x = anyRes.extract_float();
        obj = new Float(x);
        break;
      }
      case TCKind._tk_double: {
        double x = anyRes.extract_double();
        obj = new Double(x);
        break;
      }


    }

    return obj;
  };

  public  boolean isSimpleType(int typeCodeValue){
    switch(typeCodeValue){
      case org.omg.CORBA.TCKind._tk_long:
      case org.omg.CORBA.TCKind._tk_boolean:
      case org.omg.CORBA.TCKind._tk_char:
      case org.omg.CORBA.TCKind._tk_double:
      case org.omg.CORBA.TCKind._tk_float:
      case org.omg.CORBA.TCKind._tk_longdouble:
      case org.omg.CORBA.TCKind._tk_longlong:
      case org.omg.CORBA.TCKind._tk_short:
      case org.omg.CORBA.TCKind._tk_octet:
      case org.omg.CORBA.TCKind._tk_string:
      case org.omg.CORBA.TCKind._tk_ulong:
      case org.omg.CORBA.TCKind._tk_ulonglong:
      case org.omg.CORBA.TCKind._tk_ushort:
      case org.omg.CORBA.TCKind._tk_wchar:
      case org.omg.CORBA.TCKind._tk_wstring:

          return true;
      default: return false;
    }
  }
  protected  String getSimpleTypeName(int typeCodeValue){
    String result=null;
    switch(typeCodeValue){
     case org.omg.CORBA.TCKind._tk_long:
           result="int";
           break;
     case org.omg.CORBA.TCKind._tk_boolean:
       result="boolean";
           break;
     case org.omg.CORBA.TCKind._tk_char:
       result="char";
           break;
     case org.omg.CORBA.TCKind._tk_double:
       result="double";
           break;
     case org.omg.CORBA.TCKind._tk_float:
       result="float";
           break;
     case org.omg.CORBA.TCKind._tk_longdouble:
       result="long";
           break;
     case org.omg.CORBA.TCKind._tk_longlong:
       result="long";
           break;
     case org.omg.CORBA.TCKind._tk_short:
       result="short";
           break;
     case org.omg.CORBA.TCKind._tk_octet:
       result="unsigedByte";
           break;
     case org.omg.CORBA.TCKind._tk_string:
       result="string";
           break;
     case org.omg.CORBA.TCKind._tk_ulong:
       result="unsigedint";
           break;
     case org.omg.CORBA.TCKind._tk_ulonglong:
       result="unsigedlong";
           break;
     case org.omg.CORBA.TCKind._tk_ushort:
       result="unsigedshort";
           break;
     case org.omg.CORBA.TCKind._tk_wchar:
       result="wchar";
           break;
     case org.omg.CORBA.TCKind._tk_wstring:
       result="wstring";
           break;


     default: return "";
   }
   return result;
  }

  protected  Class getSimpleTypeClass(int typeCodeValue){
    Class result=null;
    switch(typeCodeValue){
     case org.omg.CORBA.TCKind._tk_long:
           result=java.lang.Integer.class;
           break;
     case org.omg.CORBA.TCKind._tk_boolean:
       result=java.lang.Boolean.class;
           break;
     case org.omg.CORBA.TCKind._tk_char:
       result=java.lang.Character.class;
           break;
     case org.omg.CORBA.TCKind._tk_double:
       result=java.lang.Double.class;
           break;
     case org.omg.CORBA.TCKind._tk_float:
       result=java.lang.Float.class;
           break;
     case org.omg.CORBA.TCKind._tk_longdouble:
       result=java.lang.Long.class;
           break;
     case org.omg.CORBA.TCKind._tk_longlong:
       result=java.lang.Long.class;
           break;
     case org.omg.CORBA.TCKind._tk_short:
       result=java.lang.Short.class;
           break;
     case org.omg.CORBA.TCKind._tk_octet:
       result=org.apache.axis.types.UnsignedByte.class;
           break;
     case org.omg.CORBA.TCKind._tk_string:
       result=String.class;
           break;
     case org.omg.CORBA.TCKind._tk_ulong:
       result=org.apache.axis.types.UnsignedInt.class;
           break;
     case org.omg.CORBA.TCKind._tk_ulonglong:
       result=org.apache.axis.types.UnsignedLong.class;
           break;
     case org.omg.CORBA.TCKind._tk_ushort:
       result=org.apache.axis.types.UnsignedShort.class;
           break;
     case org.omg.CORBA.TCKind._tk_wchar:
       result=WChar.class;
           break;
     case org.omg.CORBA.TCKind._tk_wstring:
       result=String.class;
           break;


     default: return null;
   }
   return result;
  }

  private  java.lang.Object createDynAny_fromAny(ORB _orb,int typeCode,Any anyRes) throws Exception{
    java.lang.Object result=null;
    Object objFact = null;
    try {
      objFact = _orb.resolve_initial_references("DynAnyFactory");

    }
    catch (Exception e) {
      System.out.println("error in resolve DynAnyFacotry");
      throw new Exception(e);
    }
    ;

    DynAnyFactory factory = DynAnyFactoryHelper.narrow(objFact);
    org.omg.DynamicAny.DynAny da = null;
    try {
      da = factory.create_dyn_any(anyRes);


    }
    catch (Exception e) {
      System.out.println(e.toString());
      throw new IllegalArgumentException(e);
    }
    switch(typeCode){
      case TCKind._tk_array:
        org.omg.DynamicAny.DynArray dynArray = org.omg.DynamicAny.DynArrayHelper.narrow(da) ;
        da.destroy();
        result=dynArray;
        break;
      case TCKind._tk_sequence:
         org.omg.DynamicAny.DynSequence dynSeq = org.omg.DynamicAny.DynSequenceHelper.narrow(da) ;
         da.destroy();
         result=dynSeq;
         break;
       case TCKind._tk_struct:
          org.omg.DynamicAny.DynStruct dynStruct = org.omg.DynamicAny.DynStructHelper.narrow(da) ;
          da.destroy();
          result=dynStruct;
          break;
      case TCKind._tk_union:
             org.omg.DynamicAny.DynUnion dynUnion = org.omg.DynamicAny.DynUnionHelper.narrow(da) ;
             da.destroy();
             result=dynUnion;
             break;
    case TCKind._tk_enum:
           org.omg.DynamicAny.DynEnum dynEnum=org.omg.DynamicAny.DynEnumHelper.narrow(da);
           da.destroy();
           result=dynEnum;
           break;

     case TCKind._tk_fixed:
       org.omg.DynamicAny.DynFixed dynFixed = org.omg.DynamicAny.DynFixedHelper.narrow(da);
       da.destroy();
       result = dynFixed;
       break;


    }
    return result;
  }

  public  java.lang.Object getComplexConvertArray(ORB _orb,Any anyRes,TypeCode code,String typeName){
    java.lang.Object result = null;
    try{

      org.omg.DynamicAny.DynArray dynArray = (org.omg.DynamicAny.DynArray)
          createDynAny_fromAny(_orb, code.kind().value(), anyRes);
      int num = dynArray.component_count();
      int typeValue=dynArray.current_component().type().kind().value();

        result = new ComplexType();
        String childTypeName = null;
        do {
          DynAny current = dynArray.current_component();
          if (isSimpleType(current.type().kind().value())) {
            ( (ComplexType) result).add(any2Simple(current.to_any(),
                current.type()), "item");
            childTypeName = getSimpleTypeName(current.type().kind().value());
          }
          else {

             java.lang.Object temp =  any2Complex(_orb, current.to_any(),
                current.type());
            if(temp instanceof ComplexType){
              childTypeName = ((ComplexType)temp).getNamePrefix();
              if(childTypeName.equals("")){
                childTypeName="item";
                ((ComplexType)temp).setNamePrefix(childTypeName);
              }
              ((ComplexType) result).add(temp, childTypeName);
            }
            else ((ComplexType) result).add(temp, "item");


          }
        }
        while (dynArray.next());

        if (typeName == null) {
          if (code.kind().value() == TCKind._tk_array) {
            //result.setNameFrefix("item");
            typeName = "ArrayOf" + childTypeName;
          }
          try{

            code.id();
            typeName = "_SE_" + typeName;
          }catch(Exception e){
           ((ComplexType) result).setSpecialName(true);
          }


        }

        ( (ComplexType) result).setNamePostfix(typeName);

      //}

    }catch(Exception e){
      return null;
    }

    return result;
  }

  public void setDefaultType(QName defaultType){
     this.defaultType=defaultType;
  }
}
