/***********************************************
 * Title:        Test for CORBAWebService
 * Description:  The version is developed  at June,2002
 * Copyright:    Copyright (c) 2002
 * Company:      ƴ nudt_613
 * @author       CORBA WebService Group
 * @version      1.0   
*********************************************/

package PrimitiveApp;

import org.omg.CORBA.*;
import org.omg.PortableServer.*;
import org.omg.CosNaming.*;
import org.omg.CosNaming.NamingContextPackage.*;
import java.io.*;


public class Server
{

    public static void
    main(String args[])
    {
	java.util.Properties props = System.getProperties();
	props.put("org.omg.CORBA.ORBClass", "net.nis.CORBA.ORB");
        props.put("org.omg.CORBA.ORBSingletonClass",
                  "net.nis.CORBA.ORBSingleton");

	int status = 0;
	ORB orb = null;

	try
	{
	    //
	    // Create ORB
	    //
	    orb = ORB.init(args, props);
	    status = run(orb, args);
        }
	catch(Exception ex)
	{
	    ex.printStackTrace();
	    status = 1;
	}

	if(orb != null)
	{
	    try
	    {
	        ((net.nis.CORBA.ORB)orb).destroy();
                System.out.println("The destroy");
	    }
	    catch(Exception ex)
	    {
	        System.err.println(ex.getMessage());
		ex.printStackTrace();
		status = 1;
	    }
	}

        System.exit(status);
    }

    static int
    run(ORB orb, String args[])
    {
        //
        // Parse command line arguments
        //
       String namingFile=null;
       int argc = 0;
        while(argc < args.length)
        {
	    if(args[argc].equals("-f"))
	    {
	        if(argc + 1 < args.length)
	        {
		    argc++;
		    namingFile = args[argc];
	        }
	        else
		    usage();
	    }
	    else if(args[argc].equals("-h") || args[argc].equals("--help"))
	        usage();

	    argc++;
        }

        //
        // Resolve Root POA
        //
        org.omg.CORBA.Object poaObj = null;
        try
        {
            poaObj = orb.resolve_initial_references("RootPOA");
        }
        catch(org.omg.CORBA.ORBPackage.InvalidName ex)
        {
            throw new RuntimeException();
        }

        POA rootPOA = POAHelper.narrow(poaObj);

        //
	// Get a reference to the POA manager
        //
        POAManager manager = rootPOA.the_POAManager();
        net.nis.STARPortableServer.POAManager obManager =
            net.nis.STARPortableServer.POAManagerHelper.narrow(manager);

        //
        // 10
        //
        Primitive_impl implA = new Primitive_impl(rootPOA);
        Primitive primitive= implA._this(orb);



        //
        // Get naming service
        //
        org.omg.CORBA.Object obj = null;

        if(namingFile != null)
        {
	    try
	    {
	        FileReader r = new FileReader(namingFile);
	        BufferedReader in = new BufferedReader(r);
	        String ref = in.readLine();
	        r.close();
	        obj = orb.string_to_object(ref);

	    }
	    catch(IOException e)
	    {
	        e.printStackTrace();
	        return 1;
	    }
           }
        else
         {
	    try
	    {
	        obj = orb.resolve_initial_references("NameService");
	    }
	    catch(org.omg.CORBA.ORBPackage.InvalidName ex)
	    {
	        System.out.println("Can't resolve 'NameService'");
	        return 1;
	    }
        }

        if(obj == null)
        {
	    System.out.println("`NameService' is a nil object reference");
	    return 1;
        }

        NamingContext nc = null;
        try
        {
            nc = NamingContextHelper.narrow(obj);
        }
        catch(org.omg.CORBA.BAD_PARAM ex)
        {
	    System.out.println("`NameService' is not " +
			       "a NamingContext object reference");
	    return 1;
        }

        try
        {

	    //
	    // Bind names with the Naming Service
	    //
	    NameComponent[] aName = new NameComponent[1];
	    aName[0] = new NameComponent();
	    aName[0].id = "primitive";
	    aName[0].kind = "";
	    nc.rebind(aName, primitive);



	    //
	    // Run implementation
	    //
	    try
	    {
	        manager.activate();
	    }
	    catch(org.omg.PortableServer.POAManagerPackage.AdapterInactive ex)
	    {
	        throw new RuntimeException();
	    }
	    orb.run();

	    //
	    // Unregister names with the Naming Service
	    //
        }
        catch(NotFound ex)
        {
	    System.err.print("Got a `NotFound' exception (");
	    switch(ex.why.value())
	    {
	    case NotFoundReason._missing_node:
	        System.err.print("missing node");
	        break;

	    case NotFoundReason._not_context:
	        System.err.print("not context");
	        break;

	    case NotFoundReason._not_object:
	        System.err.print("not object");
	        break;
	    }
	    System.err.println(")");
	    ex.printStackTrace();
	    return 1;
        }
        catch(CannotProceed ex)
        {
	    System.err.println("Got a `CannotProceed' exception");
	    ex.printStackTrace();
	    return 1;
        }
        catch(InvalidName ex)
        {
	    System.err.println("Got an `InvalidName' exception");
	    ex.printStackTrace();
	    return 1;
        }

	return 0;
    }

    private static void
    usage()
    {
	System.out.println(
	    "Usage: java naming.Server [options]\n\n" +
	    "Options:\n" +
	    "-f FILE     Read the NamingService IOR from file FILE.\n" +
	    "-h, --help  Display this help message.");

        System.exit(1);
    }
}
